import json
import unittest
from os.path import join as path_join

import numpy as np
from numpy.testing import assert_almost_equal
from pkg_resources import resource_filename
from pymatgen.io.vasp import Vasprun

from sumo.electronic_structure.optics import calculate_dielectric_properties, kkr


class AbsorptionTestCase(unittest.TestCase):
    def setUp(self):
        diel_path = resource_filename(
            __name__, path_join("..", "data", "Ge", "ge_diel.json")
        )
        with open(diel_path) as f:
            self.ge_diel = json.load(f)

        absorption_path = resource_filename(
            __name__, path_join("..", "data", "Ge", "ge_alpha.json")
        )
        with open(absorption_path) as f:
            self.ge_abs = json.load(f)

    def test_absorption(self):
        energy, properties = calculate_dielectric_properties(
            self.ge_diel,
            {"absorption"},
        )
        self.assertIsNone(assert_almost_equal(properties["absorption"], self.ge_abs))


class KramersKronigTestCase(unittest.TestCase):
    def setUp(self):
        ge_vasprun_path = resource_filename(
            __name__, path_join("..", "data", "Ge", "vasprun.xml.gz")
        )
        self.ge_vasprun = Vasprun(ge_vasprun_path)

        self.ge_text_file = resource_filename(
            __name__, path_join("..", "data", "Ge", "optics.txt")
        )

    def test_kkr(self):
        energy, eps_real, eps_imag = self.ge_vasprun.dielectric

        de = (energy[10] - energy[0]) / 10

        def symmetrise(a):
            """Convert XX YY ZZ XY YZ XZ array to a symmetrical 3x3 matrix"""
            return [[a[0], a[3], a[5]], [a[4], a[1], a[4]], [a[5], a[4], a[2]]]

        eps_imag_3x3 = [symmetrise(a) for a in eps_imag]
        eps_real_3x3 = np.array([symmetrise(a) for a in eps_real])

        # Check difference between eps_real reported by Vasp and determined
        # by Kramers-Kronig transformation of eps_im.
        #
        # Some discrepancy is normal, check RMS is as expected
        # This is likely due to the limited precision available in vasprun

        error = kkr(de, eps_imag_3x3) - eps_real_3x3
        error_fracs = [
            eps / eps_ref
            for eps, eps_ref in zip(error.flatten(), eps_real_3x3.flatten())
            if eps_ref > 1e-2
        ]  # Exclude low-precision cases

        self.assertLess(np.sqrt((np.array(error_fracs) ** 2).mean()), 0.1)



import unittest
import warnings
from os.path import join as path_join

import pkg_resources
from pymatgen.core.structure import Structure

from sumo.symmetry.seekpath_kpath import SeekpathKpath


class SeekpathKpathTestCase(unittest.TestCase):
    def setUp(self):
        ge_poscar = pkg_resources.resource_filename(
            __name__, path_join("..", "data", "Ge", "POSCAR")
        )
        with warnings.catch_warnings():  # Not interested in Pymatgen warnings
            warnings.simplefilter("ignore")
            self.ge_structure = Structure.from_file(ge_poscar)

    def test_seekpath_conversion(self):
        """Check path format conversion from seekpath"""
        in_path = [("A", "B"), ("B", "C"), ("D", "GAMMA")]
        in_point_coords = {
            "A": [1.0, 0.0, 0.0],
            "B": [0.0, 1.0, 0.0],
            "C": [0.0, 0.0, 1.0],
            "D": [0.5, 0.5, 0.0],
            "E": [1.0, 0.0, 1.0],
            "GAMMA": [0.0, 0.0, 0.0],
        }

        kpath = SeekpathKpath.kpath_from_seekpath(in_path, in_point_coords)
        self.assertEqual(kpath["path"], [["A", "B", "C"], ["D", r"\Gamma"]])
        self.assertEqual(
            kpath["kpoints"],
            {
                "A": [1.0, 0.0, 0.0],
                "B": [0.0, 1.0, 0.0],
                "C": [0.0, 0.0, 1.0],
                "D": [0.5, 0.5, 0.0],
                r"\Gamma": [0.0, 0.0, 0.0],
            },
        )

    def test_seekpath_kpath_load(self):
        """Check path generation from structure with seekpath"""
        kpath = SeekpathKpath(self.ge_structure)
        self.assertEqual(kpath.spg_symbol, "Fd-3m")
        self.assertEqual(kpath.lattice_type, "cubic")
        self.assertEqual(kpath.spg_number, 227)
        self.assertEqual(kpath.kpoints["L"], [0.5, 0.5, 0.5])
        self.assertEqual(
            kpath.path, [[r"\Gamma", "X", "U"], ["K", r"\Gamma", "L", "W", "X"]]
        )
        self.assertEqual(
            kpath.path_string, r"\Gamma -> X -> U | K -> \Gamma -> L -> W -> X"
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
A script to generate input files for band structure calculations

TODO:
  * Add support for CDML labels
  * Save Brillouin zone diagram
  * Return a list of filenames/folders
"""


import argparse
import logging
import os
import sys

import numpy as np
from pymatgen.io.vasp.inputs import Kpoints, Poscar

import sumo.io.questaal
import sumo.io.vasp
from sumo.io.castep import CastepCell
from sumo.io.questaal import QuestaalInit, QuestaalSite
from sumo.symmetry.kpoints import get_path_data

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "July 6, 2017"


phonon_supported_codes = {
    "castep",
}


def kgen(
    filename="POSCAR",
    code="vasp",
    directory=None,
    make_folders=False,
    symprec=0.01,
    kpts_per_split=None,
    ibzkpt=None,
    spg=None,
    density=60,
    phonon=False,
    mode="bradcrack",
    cart_coords=False,
    kpt_list=None,
    labels=None,
):
    """Generate KPOINTS files for VASP band structure calculations.

    This script provides a wrapper around several frameworks used to generate
    k-points along a high-symmetry path. The paths found in Bradley and
    Cracknell, SeeK-path, and pymatgen are all supported.

    It is important to note that the standard primitive cell symmetry is
    different between SeeK-path and pymatgen. If the correct the structure
    is not used, the high-symmetry points (and band path) may be invalid.

    Args:
        filename (:obj:`str`, optional): Path to VASP structure file. Default
            is ``POSCAR``.
        code (:obj:`str`, optional): Calculation type. Default is 'vasp';
            'questaal' also supported.
        directory (:obj:`str`, optional): The output file directory.
        make_folders (:obj:`bool`, optional): Generate folders and copy in
            required files (INCAR, POTCAR, POSCAR, and possibly CHGCAR) from
            the current directory.
        symprec (:obj:`float`, optional): The precision used for determining
            the cell symmetry.
        kpts_per_split (:obj:`int`, optional): If set, the k-points are split
            into separate k-point files (or folders) each containing the number
            of k-points specified. This is useful for hybrid band structure
            calculations where it is often intractable to calculate all
            k-points in the same calculation.
        ibzkpt (:obj:`str`, optional): Path to IBZKPT file. If set, the
            generated k-points will be appended to the k-points in this file
            and given a weight of 0. This is necessary for hybrid band
            structure calculations.
        spg (:obj:`str` or :obj:`int`, optional): The space group international
            number or symbol to override the symmetry determined by spglib.
            This is not recommended and only provided for testing purposes.
            This option will only take effect when ``mode = 'bradcrack'``.
        line_density (:obj:`int`, optional): Density of k-points along the
            path.
        phonon (:obj:`bool`, optional): Write phonon q-point path instead of
            k-points. (Not appropriate for most codes.)
        mode (:obj:`str`, optional): Method used for calculating the
            high-symmetry path. The options are:

            bradcrack
                Use the paths from Bradley and Cracknell. See [brad]_.

            pymatgen
                Use the paths from pymatgen. See [curt]_.

            latimer-munro
                Use the paths from Latimer & Munro. See [lm]_.

            seekpath
                Use the paths from SeeK-path. See [seek]_.

        cart_coords (:obj:`bool`, optional): Whether the k-points are returned
            in cartesian or reciprocal coordinates. Defaults to ``False``
            (fractional coordinates).
        kpt_list (:obj:`list`, optional): List of k-points to use, formatted as
            a list of subpaths, each containing a list of fractional k-points.
            For example::

                [ [[0., 0., 0.], [0., 0., 0.5]],
                  [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]

            Will return points along ``0 0 0 -> 0 0 1/2 | 1/2 0 0
            -> 1/2 1/2 0``
        path_labels (:obj:`list`, optional): The k-point labels. These should
            be provided as a :obj:`list` of :obj:`str` for each subpath of the
            overall path. For example::

                [ ['Gamma', 'Z'], ['X', 'M'] ]

            combined with the above example for ``kpt_list`` would indicate the
            path: Gamma -> Z | X -> M. If no labels are provided, letters from
            A -> Z will be used instead. If a label begins with '@' it will be
            concealed when plotting with sumo-bandplot.
    """
    if code.lower() == "vasp":
        structure = Poscar.from_file(filename).structure
    elif code.lower() == "questaal":
        if filename.split(".")[0] == "site":
            site = QuestaalSite.from_file(filename)
            structure = site.structure
            alat = site.alat
        else:
            structure = QuestaalInit.from_file(filename).structure
            alat = None
    elif code.lower() == "castep":
        if cart_coords:
            logging.warning(
                "Ignoring request for Cartesian coordinates: "
                "not applicable to CASTEP band structure format."
            )
            cart_coords = False
        if ibzkpt:
            logging.warning(
                'Ignoring request to use IBZKPT ("hybrid mode"), '
                "for CASTEP workflow the SCF mesh should already "
                "be set in input .cell file."
            )
        structure = CastepCell.from_file(filename).structure
    else:
        raise ValueError(f'Code "{code}" not recognized.')

    if phonon and (code.lower() not in phonon_supported_codes):
        logging.error(
            "Cannot write phonon path for {code}. "
            "Supported codes: {supported}".format(
                code=code, supported=", ".join(phonon_supported_codes)
            )
        )
        sys.exit()

    kpath, kpoints, labels = get_path_data(
        structure,
        mode=mode,
        symprec=symprec,
        kpt_list=kpt_list,
        labels=labels,
        spg=spg,
        line_density=density,
        cart_coords=cart_coords,
    )

    logging.info("\nk-point label indices:")
    for i, label in enumerate(labels):
        if label:
            logging.info(f"\t{label}: {i + 1}")

    if not kpt_list and not np.allclose(
        structure.lattice.matrix, kpath.prim.lattice.matrix
    ):
        prim_filename = f"{os.path.basename(filename)}_prim"
        if code.lower() == "questaal":
            QuestaalInit.from_structure(kpath.prim).to_file(prim_filename)
        elif code.lower() == "castep":
            CastepCell.from_structure(kpath.prim).to_file(prim_filename)

        else:
            kpath.prim.to(filename=prim_filename)

        logging.error(
            "\nWARNING: The input structure does not match the "
            "expected standard\nprimitive symmetry, the path may be "
            "incorrect! Use at your own risk.\n\nThe correct "
            "symmetry primitive structure has been saved as {}.".format(prim_filename)
        )

    ibz = _parse_ibzkpt(ibzkpt)

    if ibz and kpts_per_split is None:
        logging.info(
            "\nFound {} total kpoints in path, do you want to "
            "split them up? (y/n)".format(len(kpoints))
        )
        if input()[0].lower() == "y":
            logging.info("How many kpoints per file?")
            kpts_per_split = int(input())

    if code.lower() == "vasp":
        sumo.io.vasp.write_kpoint_files(
            filename,
            kpoints,
            labels,
            make_folders=make_folders,
            ibzkpt=ibz,
            kpts_per_split=kpts_per_split,
            directory=directory,
            cart_coords=cart_coords,
        )

    elif code.lower() == "castep":
        sumo.io.castep.write_kpoint_files(
            filename,
            kpoints,
            labels,
            make_folders=make_folders,
            kpts_per_split=kpts_per_split,
            phonon=phonon,
            directory=directory,
        )

    elif code.lower() == "questaal":
        if cart_coords:
            kpoints = [kpoint / (2 * np.pi) for kpoint in kpoints]
            if alat is not None:
                logging.info(f"Multiplying kpoint values by ALAT = {alat} Bohr")
                _bohr_to_angstrom = 0.5291772
                kpoints = [kpoint * alat * _bohr_to_angstrom for kpoint in kpoints]
        sumo.io.questaal.write_kpoint_files(
            filename,
            kpoints,
            labels,
            make_folders=make_folders,
            directory=directory,
            cart_coords=cart_coords,
        )


def _parse_ibzkpt(ibzkpt):
    if ibzkpt:
        try:
            ibz = Kpoints.from_file(ibzkpt)
            if ibz.tet_number != 0:
                logging.error("\nERROR: IBZKPT contains tetrahedron information.")
                sys.exit()
        except OSError:
            logging.error("\nERROR: Hybrid specified but no IBZKPT file found.")
            sys.exit()
    else:
        ibz = None
    return ibz


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    kgen generates KPOINTS files for running band structure calculations in
    VASP""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "-p",
        "--poscar",
        default="POSCAR",
        metavar="P",
        help="input structure file (default: POSCAR)",
    )
    parser.add_argument(
        "-c",
        "--code",
        default="vasp",
        help="Electronic structure code (default: vasp)." '"questaal" also supported.',
    )
    parser.add_argument(
        "-d",
        "--directory",
        type=str,
        default=None,
        metavar="D",
        help="output directory for files",
    )
    parser.add_argument(
        "-s",
        "--split",
        type=int,
        default=None,
        metavar="N",
        help="number of k-points to include per file",
    )
    parser.add_argument(
        "-f",
        "--folders",
        action="store_true",
        help="create folders and copy in necessary files",
    )
    parser.add_argument(
        "-y",
        "--hybrid",
        default=False,
        action="store_true",
        help="append k-points to IBZKPT file with zero weight",
    )
    parser.add_argument(
        "--symprec",
        default=0.01,
        type=float,
        help="tolerance for finding symmetry (default: 0.01)",
    )
    parser.add_argument(
        "--spg", type=str, default=None, help="space group number or symbol"
    )
    parser.add_argument(
        "--density",
        type=int,
        default=60,
        help="k-point density along high-symmetry path",
    )
    parser.add_argument(
        "--seekpath",
        action="store_true",
        help="use seekpath to generate the high-symmetry path",
    )
    parser.add_argument(
        "--pymatgen",
        action="store_true",
        help="use pymatgen to generate the high-symmetry path",
    )
    parser.add_argument(
        "--latimer-munro",
        action="store_true",
        help="use Latimer & Munro method to generate the high-symmetry path",
    )
    parser.add_argument(
        "--cartesian", action="store_true", help="use cartesian k-point coordinates"
    )
    parser.add_argument(
        "--kpoints",
        type=str,
        default=None,
        help=("specify a list of kpoints " '(e.g. "0 0 0, 0.5 0 0")'),
    )
    parser.add_argument(
        "--labels",
        type=str,
        default=None,
        help=("specify the labels for kpoints " r'(e.g. "\Gamma,X")'),
    )
    parser.add_argument(
        "--phonon",
        action="store_true",
        help="Write k-points as phonon band structure path",
    )
    return parser


def main():
    args = _get_parser().parse_args()

    logging.basicConfig(
        filename="sumo-kgen.log",
        level=logging.DEBUG,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    mode = "bradcrack"
    if args.seekpath:
        mode = "seekpath"
    elif args.pymatgen:
        mode = "pymatgen"
    elif args.latimer_munro:  # argparse converts hyphens in CLI arguments to underscore
        mode = "latimer-munro"

    ibzkpt = "IBZKPT" if args.hybrid else None

    spg = args.spg
    if args.spg:
        try:
            spg = int(spg)
        except ValueError:
            pass

    kpoints = None
    if args.kpoints:
        kpoints = [
            [list(map(float, kpt.split())) for kpt in kpts.split(",")]
            for kpts in args.kpoints.split("|")
        ]
    labels = None
    if args.labels:
        labels = [path.split(",") for path in args.labels.split("|")]

    kgen(
        args.poscar,
        code=args.code,
        directory=args.directory,
        symprec=args.symprec,
        make_folders=args.folders,
        kpts_per_split=args.split,
        ibzkpt=ibzkpt,
        spg=spg,
        density=args.density,
        mode=mode,
        cart_coords=args.cartesian,
        kpt_list=kpoints,
        labels=labels,
        phonon=args.phonon,
    )


if __name__ == "__main__":
    main()



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing helper functions for dealing with band structures.

todo:
  * Extend get projections to allow specifying lm orbitals and atomic sites.
"""

import itertools as it
from collections import defaultdict
from copy import deepcopy

import numpy as np
from pymatgen.electronic_structure.core import Spin


def get_projections_by_branches(bs, selection, normalise=None):
    """Returns orbital projections for each branch in a band structure.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structure.
        selection (list): A list of :obj:`tuple` or :obj:`string`
            identifying which projections to return. Projections can be
            specified by both element and orbital, for example::

                [('Sn', 's'), ('Bi', 'p'), ('S', 'p')]

            If just the element is specified then all the orbitals of that
            element are combined. For example, the following will combine
            all the S orbitals into a single projection::

                [('Bi', 's'), ('Bi', 'p'), 'S']

            Particular orbitals can also be combined, for example::

                [('Bi', 's'), ('Bi', 'p'), ('S', ('s', 'p', 'd'))]

        normalise (:obj:`str`, optional): Normalisation the projections.
            Options are:

              * ``'all'``: Projections normalised against the sum of all
                   other projections.
              * ``'select'``: Projections normalised against the sum of the
                   selected projections.
              * ``None``: No normalisation performed.

            Defaults to ``None``.

    Returns:
        list: A ``list`` of orbital projections for each branch of the band
        structure, in the same order as specified in ``selection``, with
        the format::

            [ [ {spin: projections} ], [ {spin: projections} ], ... ]

        Where spin is a :obj:`pymatgen.electronic_structure.core.Spin`
        object and projections is a :obj:`numpy.array` of::

            projections[band_index][kpoint_index]

        If there are no projections in the band structure, then an array of
        zeros is returned for each spin.
    """
    spins = bs.bands.keys()
    projections = get_projections(bs, selection, normalise=normalise)

    branches = []
    for b in bs.branches:
        s = b["start_index"]
        e = b["end_index"] + 1

        branch_proj = deepcopy(projections)
        for spin, i in it.product(spins, range(len(projections))):
            branch_proj[i][spin] = projections[i][spin][:, s:e]

        branches.append(branch_proj)
    return branches


def get_projections(bs, selection, normalise=None):
    """Returns orbital projections from a band structure.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structure.
        selection (list): A list of :obj:`tuple` or :obj:`string`
            identifying which projections to return. Projections can be
            specified by both element and orbital, for example::

                [('Bi', 's'), ('Bi', 'p'), ('S', 'p')]

            If just the element is specified then all the orbitals of
            that element are combined. For example, the following will combine
            all the S orbitals into a single projection::

                [('Bi', 's'), ('Bi', 'p'), 'S']

            Particular orbitals can also be combined, for example::

                [('Bi', 's'), ('Bi', 'p'), ('S', ('s', 'p', 'd'))]

        normalise (:obj:`str`, optional): Normalisation the projections.
            Options are:

              * ``'all'``: Projections normalised against the sum of all
                   other projections.
              * ``'select'``: Projections normalised against the sum of the
                   selected projections.
              * ``None``: No normalisation performed.

            Defaults to ``None``.

    Returns:
        list: A ``list`` of orbital projections, in the same order as specified
        in ``selection``, with the format::

            [ {spin: projections}, {spin: projections} ... ]

        Where spin is a :obj:`pymatgen.electronic_structure.core.Spin`
        object and projections is a :obj:`numpy.array` of::

            projections[band_index][kpoint_index]

        If there are no projections in the band structure, then an array of
        zeros is returned for each spin.
    """
    spins = bs.bands.keys()
    nbands = bs.nb_bands
    nkpts = len(bs.kpoints)

    # if we are to normalise the data later we need access to all projections
    elements = bs.structure.symbol_set
    all_orbitals = ["s", "p", "d", "f"]

    # dictio has the form: {'el1': [s, p, d, f], 'el2': [s, p, d, f]...}
    dictio = dict(zip(elements, [all_orbitals] * len(elements)))

    # bs.get_projection_on_elements_and_orbitals() returns the data in a
    # really frustrating format, namely:
    #     {spin: [band_index][kpoint_index]{element: {orbital: projection}}}
    all_proj = bs.get_projections_on_elements_and_orbitals(dictio)

    # Make a defaultdict of defaultdicts
    dict_proj = defaultdict(lambda: defaultdict(dict))
    sum_proj = {s: np.zeros((nbands, nkpts)) for s in spins}

    # store the projections for all elements and orbitals in a useable format
    for spin, element, orbital in it.product(spins, elements, all_orbitals):

        # convert data to [nb][nk]
        el_orb_proj = [
            [all_proj[spin][nb][nk][element][orbital] for nk in range(nkpts)]
            for nb in range(nbands)
        ]

        dict_proj[element][orbital][spin] = np.array(el_orb_proj)

        if normalise == "all":
            sum_proj[spin] += el_orb_proj

    # now go through the selected orbitals and extract what's needed
    spec_proj = []
    for spec in selection:

        if isinstance(spec, str):
            # spec is just an element type, therefore sum all orbitals
            element = spec
            orbitals = all_orbitals
        else:
            element, orbitals = spec
            # even if there is only one orbital, make sure we can loop over it
            orbitals = tuple(orbitals)

        proj = {s: np.zeros((nbands, nkpts)) for s in spins}
        for spin, orbital in it.product(spins, orbitals):
            proj[spin] += dict_proj[element][orbital][spin]

            if normalise == "select":
                sum_proj[spin] += dict_proj[element][orbital][spin]

        spec_proj.append(proj)

    if normalise:
        # to prevent warnings/errors relating to divide by zero,
        # catch warnings and surround divide with np.nan_to_num
        with np.errstate(divide="ignore", invalid="ignore"):
            for spin, i in it.product(spins, range(len(spec_proj))):
                spec_proj[i][spin] = np.nan_to_num(spec_proj[i][spin] / sum_proj[spin])
    return spec_proj


def get_reconstructed_band_structure(list_bs, efermi=None, force_kpath_branches=True):
    """Combine a list of band structures into a single band structure.

    This is typically very useful when you split non self consistent
    band structure runs in several independent jobs and want to merge back
    the results.

    This method will also ensure that any BandStructure objects will contain
    branches.

    Args:
        list_bs (:obj:`list` of \
        :obj:`~pymatgen.electronic_structure.bandstructure.BandStructure` \
        or :obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structures.
        efermi (:obj:`float`, optional): The Fermi energy of the reconstructed
            band structure. If `None`, an average of all the Fermi energies
            across all band structures is used.
        force_kpath_branches (bool): Force a linemode band structure to contain
            branches by adding repeated high-symmetry k-points in the path.

    Returns:
        :obj:`pymatgen.electronic_structure.bandstructure.BandStructure` or \
        :obj:`pymatgen.electronic_structure.bandstructureBandStructureSymmLine`:
        A band structure object. The type depends on the type of the band
        structures in ``list_bs``.
    """
    if efermi is None:
        efermi = sum(b.efermi for b in list_bs) / len(list_bs)

    kpoints = np.concatenate([[k.frac_coords for k in bs.kpoints] for bs in list_bs])
    nb_bands = min(list_bs[i].nb_bands for i in range(len(list_bs)))
    dicts = [bs.labels_dict for bs in list_bs]
    labels_dict = {k: v.frac_coords for d in dicts for k, v in d.items()}

    eigenvals = {}
    projections = {}
    for spin in list_bs[0].bands.keys():
        spin_bands = [bs.bands[spin][:nb_bands] for bs in list_bs]
        eigenvals[spin] = np.concatenate(spin_bands, axis=1)

        if len(list_bs[0].projections) != 0:
            spin_projs = [bs.projections[Spin.up][:nb_bands] for bs in list_bs]
            projections[Spin.up] = np.concatenate(spin_projs, axis=1)

    bs = type(list_bs[0])(
        kpoints,
        eigenvals,
        list_bs[0].lattice_rec,
        efermi,
        labels_dict,
        structure=list_bs[0].structure,
        projections=projections,
    )
    if force_kpath_branches:
        return force_branches(bs)
    else:
        return bs


def force_branches(bandstructure):
    """Force a linemode band structure to contain branches.

    Branches give a specific portion of the path from one high-symmetry point
    to another. Branches are required for the plotting methods to function correctly.
    Unfortunately, due to the pymatgen BandStructure implementation they require
    duplicate k-points in the band structure path. To avoid this unnecessary
    computational expense, this function can reconstruct branches in band structures
    without the duplicate k-points.

    Args:
        bandstructure: A band structure object.

    Returns:
        A band structure with brnaches.
    """
    kpoints = np.array([k.frac_coords for k in bandstructure.kpoints])
    labels_dict = {k: v.frac_coords for k, v in bandstructure.labels_dict.items()}

    # pymatgen band structure objects support branches. These are formed when
    # two kpoints with the same label are next to each other. This bit of code
    # will ensure that the band structure will contain branches, if it doesn't
    # already.
    dup_ids = []
    high_sym_kpoints = tuple(map(tuple, labels_dict.values()))
    for i, k in enumerate(kpoints):
        dup_ids.append(i)
        if (
            tuple(k) in high_sym_kpoints
            and i != 0
            and i != len(kpoints) - 1
            and (
                not np.array_equal(kpoints[i + 1], k)
                or not np.array_equal(kpoints[i - 1], k)
            )
        ):
            dup_ids.append(i)

    kpoints = kpoints[dup_ids]

    eigenvals = {}
    projections = {}
    for spin, spin_energies in bandstructure.bands.items():
        eigenvals[spin] = spin_energies[:, dup_ids]
        if len(bandstructure.projections) != 0:
            projections[spin] = bandstructure.projections[spin][:, dup_ids]

    return type(bandstructure)(
        kpoints,
        eigenvals,
        bandstructure.lattice_rec,
        bandstructure.efermi,
        labels_dict,
        structure=bandstructure.structure,
        projections=projections,
    )


def string_to_spin(spin_string):
    """Function to convert 'spin' cli argument to pymatgen Spin object"""
    if spin_string in ["up", "Up", "1", "+1"]:
        return Spin.up

    elif spin_string in ["down", "Down", "-1"]:
        return Spin.down

    elif spin_string is None:
        return None

    else:
        raise ValueError("Unable to parse 'spin' argument")



import unittest

from sumo.plotting.bs_plotter import SBSPlotter


class SanitiseLabelTestCase(unittest.TestCase):
    def test_sanitise_label(self):
        for label_in, label_out in (
            ("X", "X"),
            ("X 1", "X 1"),
            ("X @", "X "),
            ("X@", "X"),
            ("X@@@", "X"),
            ("HEX", "HEX"),
            ("HE@X", "HE@X"),
            ("HE@X 1", "HE@X 1"),
            ("@", None),
            ("@X", None),
            ("@HEX", None),
        ):

            self.assertEqual(SBSPlotter._sanitise_label(label_in), label_out)

    def test_sanitise_label_group(self):
        for label_in, label_out in (
            ("X", "X"),
            ("X 1", "X 1"),
            ("X @", "X "),
            ("X@", "X"),
            ("X@@@", "X"),
            ("HEX", "HEX"),
            ("HE@X", "HE@X"),
            ("HE@X 1", "HE@X 1"),
            ("@", None),
            ("@X", None),
            ("@HEX", None),
            (r"X$\mid$Y", r"X$\mid$Y"),
            (r"X$\mid$Y$\mid$Z", r"X$\mid$Y$\mid$Z"),
            (r"@X$\mid$Y$\mid$Z", r"Y$\mid$Z"),
            (r"X$\mid$@Y$\mid$Z", r"X$\mid$Z"),
            (r"X$\mid$@Z", r"X"),
            (r"X@$\mid$Y", r"X$\mid$Y"),
            (r"X@$\mid$Y@@", r"X$\mid$Y"),
            (r"@X@$\mid$Y@", r"Y"),
            (r"X@$\mid$@Y", r"X"),
            (r"@X@$\mid$@Y", None),
        ):

            self.assertEqual(SBSPlotter._sanitise_label_group(label_in), label_out)



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
A script to plot density of states diagrams.

TODO:
    * Add ability to scale an orbitals density of states
"""


import argparse
import logging
import os
import sys
import warnings
from glob import glob

import matplotlib as mpl
import numpy as np
from pkg_resources import Requirement, resource_filename

mpl.use("Agg")

import sumo.io.castep
import sumo.io.questaal
from sumo.electronic_structure.bandstructure import string_to_spin
from sumo.electronic_structure.dos import load_dos, write_files
from sumo.plotting.dos_plotter import SDOSPlotter

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "April 9, 2018"


def dosplot(
    filename=None,
    code="vasp",
    prefix=None,
    directory=None,
    elements=None,
    lm_orbitals=None,
    atoms=None,
    spin=None,
    subplot=False,
    shift=True,
    total_only=False,
    plot_total=True,
    legend_on=True,
    legend_frame_on=False,
    legend_cutoff=3.0,
    gaussian=None,
    colours=None,
    height=6.0,
    width=8.0,
    xmin=-6.0,
    xmax=6.0,
    num_columns=2,
    xlabel="Energy (eV)",
    ylabel="DOS",
    yscale=1,
    zero_energy=None,
    zero_line=False,
    style=None,
    no_base_style=False,
    image_format="pdf",
    dpi=400,
    plt=None,
    fonts=None,
):
    """A script to plot the density of states from a vasprun.xml file.

    Args:
        filename (:obj:`str`, optional): Path to a DOS data file (can be
            gzipped). The preferred file type depends on the electronic
            structure code: vasprun.xml (VASP); *.bands (CASTEP); dos.*
            (Questaal).
        code (:obj:`str`, optional): Electronic structure code used ('vasp',
              'castep' or 'questaal'). Note that for Castep only a rough TDOS
              is available, assembled by sampling the eigenvalues.
        prefix (:obj:`str`, optional): Prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the POSCAR. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.
        spin (:obj:`Spin`, optional): Plot only one spin channel from a
            spin-polarised calculation; "up" or "1" for spin up only, "down" or
            "-1" for spin down only. Defaults to ``None``.
        subplot (:obj:`bool`, optional): Plot the density of states for each
            element on separate subplots. Defaults to ``False``.
        shift (:obj:`bool`, optional): Shift the energies such that the valence
            band maximum (or Fermi level for metals) is at 0 eV. Defaults to
            ``True``.
        total_only (:obj:`bool`, optional): Only extract the total density of
            states. Defaults to ``False``.
        plot_total (:obj:`bool`, optional): Plot the total density of states.
            Defaults to ``True``.
        legend_on (:obj:`bool`, optional): Plot the graph legend. Defaults
            to ``True``.
        legend_frame_on (:obj:`bool`, optional): Plot a frame around the
            graph legend. Defaults to ``False``.
        legend_cutoff (:obj:`float`, optional): The cut-off (in % of the
            maximum density of states within the plotting range) for an
            elemental orbital to be labelled in the legend. This prevents
            the legend from containing labels for orbitals that have very
            little contribution in the plotting range.
        gaussian (:obj:`float`, optional): Broaden the density of states using
            convolution with a gaussian function. This parameter controls the
            sigma or standard deviation of the gaussian distribution.
        height (:obj:`float`, optional): The height of the plot.
        width (:obj:`float`, optional): The width of the plot.
        xmin (:obj:`float`, optional): The minimum energy on the x-axis.
        xmax (:obj:`float`, optional): The maximum energy on the x-axis.
        num_columns (:obj:`int`, optional): The number of columns in the
            legend.
        colours (:obj:`dict`, optional): Use custom colours for specific
            element and orbital combinations. Specified as a :obj:`dict` of
            :obj:`dict` of the colours. For example::

                {
                    'Sn': {'s': 'r', 'p': 'b'},
                    'O': {'s': '#000000'}
                }

            The colour can be a hex code, series of rgb value, or any other
            format supported by matplotlib.
        xlabel (:obj:`str`, optional): Label/units for x-axis (i.e. energy)
        ylabel (:obj:`str`, optional): Label/units for y-axis (i.e. DOS)
        yscale (:obj:`float`, optional): Scaling factor for the y-axis.
        zero_line (:obj:`bool`, optional): Plot vertical line at energy zero.
        zero_energy (:obj:`float`, optional): Zero energy reference (e.g. Fermi
            energy from sc-fermi.) If not given, behaviour is determined by
            boolean ``shift``.
        style (:obj:`list` or :obj:`str`, optional): (List of) matplotlib style
            specifications, to be composed on top of Sumo base style.
        no_base_style (:obj:`bool`, optional): Prevent use of sumo base style.
            This can make alternative styles behave more predictably.
        image_format (:obj:`str`, optional): The image file format. Can be any
            format supported by matplotlib, including: png, jpg, pdf, and svg.
            Defaults to pdf.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the image.
        plt (:obj:`matplotlib.pyplot`, optional): A
            :obj:`matplotlib.pyplot` object to use for plotting.
        fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
            a single font, specified as a :obj:`str`, or several fonts,
            specified as a :obj:`list` of :obj:`str`.

    Returns:
        A matplotlib pyplot object.
    """

    if code.lower() == "vasp":
        if not filename:
            if os.path.exists("vasprun.xml"):
                filename = "vasprun.xml"
            elif os.path.exists("vasprun.xml.gz"):
                filename = "vasprun.xml.gz"
            else:
                logging.error("ERROR: No vasprun.xml found!")
                sys.exit()

        dos, pdos = load_dos(
            filename, elements, lm_orbitals, atoms, gaussian, total_only
        )

    elif code.lower() == "castep":

        if filename:
            bands_file = filename
        else:
            band_candidates = glob("*.bands")
            if len(band_candidates) == 0:
                logging.error("ERROR: No *.bands file found!")
                sys.exit()
            elif len(band_candidates) == 1:
                bands_file = band_candidates[0]
            else:
                logging.error("ERROR: Too many *.bands files found!")
                sys.exit()
        pdos_file = _replace_ext(bands_file, "pdos_bin")
        cell_file = _replace_ext(bands_file, "cell")
        pdos_file = pdos_file if os.path.isfile(pdos_file) else None
        cell_file = cell_file if os.path.isfile(cell_file) else None

        if not total_only:
            # Check if both pdos_bin and cell files are present.
            # If not, we cannot plot the PDOS
            if pdos_file is not None:
                if cell_file is None:
                    logging.info(
                        "Plotting PDOS requires the .cell file to be "
                        "present; falling back to TDOS."
                    )
                    pdos_file = None
                else:
                    logging.info(
                        f"Found PDOS binary file {pdos_file}; "
                        "including PDOS in the plot."
                    )
            else:
                logging.info("PDOS not available, falling back to TDOS.")

        dos, pdos = sumo.io.castep.read_dos(
            bands_file,
            pdos_file=pdos_file,
            cell_file=cell_file,
            gaussian=gaussian,
            emin=xmin,
            emax=xmax,
            lm_orbitals=lm_orbitals,
            elements=elements,
            total_only=total_only,
            atoms=atoms,
        )

    elif code.lower() == "questaal":
        if filename:
            pdos_file = filename
            ext = pdos_file.split(".")[-1]
        else:
            pdos_candidates = glob("dos.*")
            for candidate in pdos_candidates:
                if candidate.split(".")[-1] in ("pdf", "png", "svg", "jpg", "jpeg"):
                    continue
                elif candidate.split(".")[-1].lower() in ("gz", "z", "bz2"):
                    pdos_file = candidate
                    ext = candidate.split(".")[-2]
                    break
                else:
                    pdos_file = candidate
                    ext = candidate.split(".")[-1]
                    break
            else:
                raise ValueError("No questaal dos file found")

        if os.path.exists(f"tdos.{ext}"):
            tdos_file = f"tdos.{ext}"
        else:
            tdos_file = None
        if os.path.exists(f"site.{ext}"):
            site_file = f"site.{ext}"
        else:
            site_file = None

        if shift:
            logging.warning(
                "Fermi level shift requested, but not implemented for Questaal DOS."
            )

        dos, pdos = sumo.io.questaal.read_dos(
            pdos_file=pdos_file,
            tdos_file=tdos_file,
            site_file=site_file,
            ry=True,
            gaussian=gaussian,
            total_only=total_only,
            elements=elements,
            lm_orbitals=lm_orbitals,
            atoms=atoms,
        )

    else:
        logging.error(f"ERROR: Unrecognised code: {code}")
        return

    save_files = False if plt else True  # don't save if pyplot object provided

    spin = string_to_spin(spin)  # Convert spin name to pymatgen Spin object
    plotter = SDOSPlotter(dos, pdos)

    plt = plotter.get_plot(
        subplot=subplot,
        width=width,
        height=height,
        xmin=xmin,
        xmax=xmax,
        yscale=yscale,
        colours=colours,
        plot_total=plot_total,
        legend_on=legend_on,
        num_columns=num_columns,
        legend_frame_on=legend_frame_on,
        xlabel=xlabel,
        ylabel=ylabel,
        zero_line=zero_line,
        zero_to_efermi=shift,
        zero_energy=zero_energy,
        legend_cutoff=legend_cutoff,
        dpi=dpi,
        plt=plt,
        fonts=fonts,
        style=style,
        no_base_style=no_base_style,
        spin=spin,
    )

    if save_files:
        basename = f"dos.{image_format}"
        filename = f"{prefix}_{basename}" if prefix else basename
        if directory:
            filename = os.path.join(directory, filename)
        plt.savefig(filename, format=image_format, dpi=dpi, bbox_inches="tight")
        write_files(dos, pdos, prefix=prefix, directory=directory)
    else:
        return plt


def _el_orb(string):
    """Parse the element and orbital argument strings.

    The presence of an element without any orbitals means that we want to plot
    all of its orbitals.

    Args:
        string (str): The element and orbitals as a string, in the form
            ``"C.s.p,O"``.

    Returns:
        dict: The elements and orbitals as a :obj:`dict`. For example::

            {'Bi': ['s', 'px', 'py', 'd']}.

        If an element symbol is included with an empty list, then all orbitals
        for that species are considered.
    """
    el_orbs = {}
    for split in string.split(","):
        orbs = split.split(".")
        orbs = [orbs[0], "s", "p", "d", "f"] if len(orbs) == 1 else orbs
        el_orbs[orbs.pop(0)] = orbs
    return el_orbs


def _replace_ext(string, new_ext):
    """Replace file extension

    Args:
        string (`str`): The file name with extensions to be replace
        new_ext (`str`): The new extension

    Returns:
        A string with files extension replaced by new_ext
    """
    name, _ = os.path.splitext(string)
    return name + "." + new_ext


def _atoms(atoms_string):
    """Parse the atom string.

    Args:
        atoms_string (str): The atoms to plot, in the form ``"C.1.2.3,"``.

    Returns:
        dict: The atomic indices over which to sum the DOS. Formatted as::

            {Element: [atom_indices]}.

        Indices are zero indexed for each atomic species. If an element symbol
        is included with an empty list, then all sites for that species are
        considered.
    """
    atoms = {}
    for split in atoms_string.split(","):
        sites = split.split(".")
        el = sites.pop(0)
        sites = list(map(int, sites))
        atoms[el] = np.array(sites) - 1
    return atoms


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    dosplot is a script to produce publication-ready density of
    states diagrams""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "-f", "--filename", help="vasprun.xml file to plot", default=None, metavar="F"
    )
    parser.add_argument(
        "-c",
        "--code",
        default="vasp",
        metavar="C",
        help=('Input file format: "vasp" (vasprun.xml) or ' '"questaal" (opt.ext)'),
    )
    parser.add_argument(
        "-p", "--prefix", metavar="P", help="prefix for the files generated"
    )
    parser.add_argument(
        "-d", "--directory", metavar="D", help="output directory for files"
    )
    parser.add_argument(
        "-e",
        "--elements",
        type=_el_orb,
        metavar="E",
        help='elemental orbitals to plot (e.g. "C.s.p,O")',
    )
    parser.add_argument(
        "-o",
        "--orbitals",
        type=_el_orb,
        metavar="O",
        help=("orbitals to split into lm-decomposed " 'contributions (e.g. "Ru.d")'),
    )
    parser.add_argument(
        "-a",
        "--atoms",
        type=_atoms,
        metavar="A",
        help=('atoms to include (e.g. "O.1.2.3,Ru.1.2.3")'),
    )
    parser.add_argument(
        "--spin",
        type=str,
        default=None,
        help=(
            "select one spin channel only for a "
            "spin-polarised calculation "
            "(options: up, 1; down, -1)"
        ),
    )
    parser.add_argument(
        "-s",
        "--subplot",
        action="store_true",
        help="plot each element on separate subplots",
    )
    parser.add_argument(
        "-g",
        "--gaussian",
        type=float,
        metavar="G",
        help="standard deviation of gaussian broadening",
    )
    parser.add_argument(
        "--columns",
        type=int,
        default=2,
        metavar="N",
        help="number of columns in the legend",
    )
    parser.add_argument(
        "--legend-cutoff",
        type=float,
        default=3,
        dest="legend_cutoff",
        metavar="C",
        help=(
            "cut-off in %% of total DOS that determines if"
            " a line is given a label (default: 3)"
        ),
    )
    parser.add_argument(
        "--no-legend", action="store_false", dest="legend", help="hide the plot legend"
    )
    parser.add_argument(
        "--legend-frame",
        action="store_true",
        dest="legend_frame",
        help="display a box around the legend",
    )

    shift = parser.add_mutually_exclusive_group()

    shift.add_argument(
        "--no-shift",
        action="store_false",
        dest="shift",
        help="don't shift the VBM/Fermi level to 0 eV",
    )
    shift.add_argument(
        "--zero-energy",
        type=float,
        default=None,
        dest="zero_energy",
        help="Plot vertical line at energy zero",
    )

    parser.add_argument(
        "--total-only",
        action="store_true",
        dest="total_only",
        help="only plot the total density of states",
    )
    parser.add_argument(
        "--no-total",
        action="store_false",
        dest="total",
        help="don't plot the total density of states",
    )
    parser.add_argument(
        "--height", type=float, default=None, help="height of the graph"
    )
    parser.add_argument("--width", type=float, default=None, help="width of the graph")
    parser.add_argument(
        "--xmin", type=float, default=-6.0, help="minimum energy on the x-axis"
    )
    parser.add_argument(
        "--xmax", type=float, default=6.0, help="maximum energy on the x-axis"
    )
    parser.add_argument(
        "--config", type=str, default=None, help="colour configuration file"
    )
    parser.add_argument(
        "--style",
        type=str,
        nargs="+",
        default=None,
        help="matplotlib style specifications",
    )
    parser.add_argument(
        "--no-base-style",
        action="store_true",
        dest="no_base_style",
        help="prevent use of sumo base style",
    )
    parser.add_argument(
        "--xlabel",
        type=str,
        default="Energy (eV)",
        help="x-axis (i.e. energy) label/units",
    )
    parser.add_argument(
        "--ylabel", type=str, default="DOS", help="y-axis (i.e. DOS) label/units"
    )
    parser.add_argument(
        "--yscale", type=float, default=1, help="scaling factor for the y-axis"
    )
    parser.add_argument(
        "--zero-line",
        action="store_true",
        dest="zero_line",
        help="Plot vertical line at energy zero",
    )
    parser.add_argument(
        "--format",
        type=str,
        default="pdf",
        dest="image_format",
        metavar="FORMAT",
        help="image file format (options: pdf, svg, jpg, png)",
    )
    parser.add_argument(
        "--dpi", type=int, default=400, help="pixel density for image file"
    )
    parser.add_argument("--font", default=None, help="font to use")
    return parser


def main():
    args = _get_parser().parse_args()
    logging.basicConfig(
        filename="sumo-dosplot.log",
        level=logging.INFO,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    if args.config is None:
        config_path = resource_filename(
            Requirement.parse("sumo"), "sumo/plotting/orbital_colours.conf"
        )
    else:
        config_path = args.config
    colours = configparser.ConfigParser()
    colours.read(os.path.abspath(config_path))

    warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UnicodeWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")

    if args.zero_energy is not None:
        shift = False
    else:
        shift = args.shift

    dosplot(
        filename=args.filename,
        code=args.code,
        prefix=args.prefix,
        directory=args.directory,
        elements=args.elements,
        lm_orbitals=args.orbitals,
        atoms=args.atoms,
        spin=args.spin,
        subplot=args.subplot,
        shift=shift,
        total_only=args.total_only,
        plot_total=args.total,
        legend_on=args.legend,
        legend_frame_on=args.legend_frame,
        legend_cutoff=args.legend_cutoff,
        gaussian=args.gaussian,
        height=args.height,
        width=args.width,
        xmin=args.xmin,
        xmax=args.xmax,
        num_columns=args.columns,
        colours=colours,
        style=args.style,
        no_base_style=args.no_base_style,
        xlabel=args.xlabel,
        ylabel=args.ylabel,
        yscale=args.yscale,
        zero_line=args.zero_line,
        zero_energy=args.zero_energy,
        image_format=args.image_format,
        dpi=args.dpi,
        fonts=args.font,
    )


if __name__ == "__main__":
    main()



import unittest
import warnings
from os.path import join as path_join

import pkg_resources
from pymatgen.core.structure import Structure

from sumo.symmetry.brad_crack_kpath import BradCrackKpath
from sumo.symmetry.pymatgen_kpath import PymatgenKpath


class SeekpathKpathTestCase(unittest.TestCase):
    def setUp(self):
        zno_poscar = pkg_resources.resource_filename(
            __name__, path_join("..", "data", "ZnO", "POSCAR")
        )

        hgs_poscar = pkg_resources.resource_filename(
            __name__, path_join("..", "data", "Ge", "POSCAR")
        )

        with warnings.catch_warnings():  # Not interested in Pymatgen warnings
            warnings.simplefilter("ignore")
            self.zno_structure = Structure.from_file(zno_poscar)
            self.hgs_structure = Structure.from_file(hgs_poscar)

    def test_pymatgen_path(self):
        """Check kpoint sequence from pymatgen: Ge example"""
        kpath = PymatgenKpath(self.zno_structure)
        self.assertEqual(
            kpath.path,
            [
                [r"\Gamma", "M", "K", r"\Gamma", "A", "L", "H", "A"],
                ["L", "M"],
                ["K", "H"],
            ],
        )

    def test_pymatgen_points(self):
        """Check special points agree between pymatgen and Bradley-Cracknell"""
        kpath_pymatgen = PymatgenKpath(self.hgs_structure)
        kpath_bradcrack = BradCrackKpath(self.hgs_structure)

        # Bradcrack kpoints should be a subset of pymatgen kpoints
        for label, position in kpath_bradcrack.kpoints.items():
            self.assertIn(label, kpath_pymatgen.kpoints)
            self.assertEqual(list(position), list(kpath_pymatgen.kpoints[label]))



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing class for generating k-points along paths from pymatgen.
"""

from pymatgen.symmetry.bandstructure import HighSymmKpath

from sumo.symmetry import Kpath


class PymatgenKpath(Kpath):
    r"""Class to generate k-points along paths from pymatgen.

    More detail on the paths generated by can be found in the pymatgen
    documentation. These paths are based on the work described in reference
    [curt]_.

    .. [curt] Setyawan, W., & Curtarolo, S. High-throughput electronic band
              structure calculations: Challenges and tools, Computational
              Materials Science, 49, 299-312 (2010). doi:
              10.1016/j.commatsci.2010.05.010

    These paths should be used with primitive structures that comply with the
    definition from the paper. This structure can be accessed using the
    ``prim`` attribute and compliance between the provided structure and
    standardised structure checked using the ``correct_structure()`` method.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.

    Attributes:
        prim (:obj:`~pymatgen.core.structure.Structure`): The standardised
            primitive cell structure for the generated k-point path.
        conv (:obj:`~pymatgen.core.structure.Structure`): The standardised
            conventional cell structure.
    """

    def __init__(self, structure, symprec=1e-3):
        Kpath.__init__(self, structure, symprec=symprec)
        pmg_path = HighSymmKpath(structure, symprec=symprec)
        self._kpath = pmg_path._kpath
        self.prim = pmg_path.prim
        self.conv = pmg_path.conventional



"""
A script to plot electronic band structure diagrams.

TODO:
 - Replace the elements and project formats with the dream syntax
"""


import argparse
import glob
import logging
import os
import sys
import warnings

import matplotlib as mpl
from pkg_resources import Requirement, resource_filename
from pymatgen.electronic_structure.bandstructure import get_reconstructed_band_structure
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp.outputs import BSVasprun

mpl.use("Agg")

from sumo.cli.dosplot import _atoms, _el_orb
from sumo.electronic_structure.bandstructure import string_to_spin
from sumo.electronic_structure.dos import load_dos
from sumo.io.castep import band_structure as castep_band_structure
from sumo.io.castep import read_dos as read_castep_dos
from sumo.io.questaal import QuestaalSite
from sumo.io.questaal import band_structure as questaal_band_structure
from sumo.io.questaal import labels_from_syml
from sumo.plotting.bs_plotter import SBSPlotter
from sumo.plotting.dos_plotter import SDOSPlotter

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "July 18, 2017"


def bandplot(
    filenames=None,
    code="vasp",
    prefix=None,
    directory=None,
    vbm_cbm_marker=False,
    projection_selection=None,
    mode="rgb",
    normalise="all",
    interpolate_factor=4,
    color1="#FF0000",
    color2="#00FF00",
    color3="#0000FF",
    colorspace="lab",
    circle_size=150,
    dos_file=None,
    cart_coords=False,
    scissor=None,
    ylabel="Energy (eV)",
    dos_label=None,
    zero_line=False,
    zero_energy=None,
    elements=None,
    lm_orbitals=None,
    atoms=None,
    spin=None,
    total_only=False,
    plot_total=True,
    legend_cutoff=3,
    gaussian=None,
    height=None,
    width=None,
    ymin=-6.0,
    ymax=6.0,
    colours=None,
    yscale=1,
    style=None,
    no_base_style=False,
    image_format="pdf",
    dpi=400,
    plt=None,
    fonts=None,
):
    """Plot electronic band structure diagrams from vasprun.xml files.

    Args:
        filenames (:obj:`str` or :obj:`list`, optional): Path to input files:

            Vasp:
                Use vasprun.xml or vasprun.xml.gz file.
            Questaal:
                Path to a bnds.ext file. The extension will also be used to
                find site.ext and syml.ext files in the same directory.
            Castep:
                Path to a seedname.bands file. The prefix ("seedname") is used
                to locate a seedname.cell file in the same directory and read
                in the positions of high-symmetry points.

            If no filenames are provided, sumo
            will search for vasprun.xml or vasprun.xml.gz files in folders
            named 'split-0*'. Failing that, the code will look for a vasprun in
            the current directory. If a :obj:`list` of vasprun files is
            provided, these will be combined into a single band structure.

        code (:obj:`str`, optional): Calculation type. Default is 'vasp';
            'questaal' and 'castep' also supported (with a reduced
            feature-set).
        prefix (:obj:`str`, optional): Prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
        vbm_cbm_marker (:obj:`bool`, optional): Plot markers to indicate the
            VBM and CBM locations.
        projection_selection (list): A list of :obj:`tuple` or :obj:`string`
            identifying which elements and orbitals to project on to the
            band structure. These can be specified by both element and
            orbital, for example, the following will project the Bi s, p
            and S p orbitals::

                [('Bi', 's'), ('Bi', 'p'), ('S', 'p')]

            If just the element is specified then all the orbitals of
            that element are combined. For example, to sum all the S
            orbitals::

                [('Bi', 's'), ('Bi', 'p'), 'S']

            You can also choose to sum particular orbitals by supplying a
            :obj:`tuple` of orbitals. For example, to sum the S s, p, and
            d orbitals into a single projection::

                [('Bi', 's'), ('Bi', 'p'), ('S', ('s', 'p', 'd'))]

            If ``mode = 'rgb'``, a maximum of 3 orbital/element
            combinations can be plotted simultaneously (one for red, green
            and blue), otherwise an unlimited number of elements/orbitals
            can be selected.
        mode (:obj:`str`, optional): Type of projected band structure to
            plot. Options are:

                "rgb"
                    The band structure line color depends on the character
                    of the band. Each element/orbital contributes either
                    red, green or blue with the corresponding line colour a
                    mixture of all three colours. This mode only supports
                    up to 3 elements/orbitals combinations. The order of
                    the ``selection`` :obj:`tuple` determines which colour
                    is used for each selection.
                "stacked"
                    The element/orbital contributions are drawn as a
                    series of stacked circles, with the colour depending on
                    the composition of the band. The size of the circles
                    can be scaled using the ``circle_size`` option.

        normalise (:obj:`str`, optional): Normalisation the projections.
            Options are:

              * ``'all'``: Projections normalised against the sum of all
                   other projections.
              * ``'select'``: Projections normalised against the sum of the
                   selected projections.
              * ``None``: No normalisation performed.

        color1 (str): A color specified in any way supported by matplotlib. Used
            when ``mode = 'rgb'``.
        color2 (str): A color specified in any way supported by matplotlib. Used
            when ``mode = 'rgb'``.
        color3 (str): A color specified in any way supported by matplotlib. Used
            when ``mode = 'rgb'``.
        colorspace (str): The colorspace in which to perform the interpolation. The
            allowed values are rgb, hsv, lab, luvlc, lablch, and xyz. Used
            when ``mode = 'rgb'``.
        circle_size (:obj:`float`, optional): The area of the circles used
            when ``mode = 'stacked'``.
        cart_coords (:obj:`bool`, optional): Whether the k-points are read as
            cartesian or reciprocal coordinates. This is only required for
            Questaal output; Vasp output is less ambiguous. Defaults to
            ``False`` (fractional coordinates).
        scissor (:obj:`float`, optional): Apply a scissor operator (rigid shift
            of the CBM), use with caution if applying to metals.
        dos_file (:obj:`str`, optional): Path to vasprun.xml file from which to
            read the density of states information. If set, the density of
            states will be plotted alongside the bandstructure.
        zero_line (:obj:`bool`, optional): If true, draw a horizontal line at
            zero energy. (To adjust where zero energy sits, use zero_energy.)
        zero_energy (:obj:`float`, optional): Energy offset determining position
            of zero energy. By default, this is the VBM. It may be useful to
            set this to e.g. a calculated self-consistent Fermi energy.
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the POSCAR. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.
        spin (:obj:`Spin`, optional): Plot only one spin channel from a
            spin-polarised calculation; "up" or "1" for spin up only, "down" or
            "-1" for spin down only. Defaults to ``None``.
        total_only (:obj:`bool`, optional): Only extract the total density of
            states. Defaults to ``False``.
        plot_total (:obj:`bool`, optional): Plot the total density of states.
            Defaults to ``True``.
        legend_cutoff (:obj:`float`, optional): The cut-off (in % of the
            maximum density of states within the plotting range) for an
            elemental orbital to be labelled in the legend. This prevents
            the legend from containing labels for orbitals that have very
            little contribution in the plotting range.
        gaussian (:obj:`float`, optional): Broaden the density of states using
            convolution with a gaussian function. This parameter controls the
            sigma or standard deviation of the gaussian distribution.
        height (:obj:`float`, optional): The height of the plot.
        width (:obj:`float`, optional): The width of the plot.
        ymin (:obj:`float`, optional): The minimum energy on the y-axis.
        ymax (:obj:`float`, optional): The maximum energy on the y-axis.
        style (:obj:`list` or :obj:`str`, optional): (List of) matplotlib style
            specifications, to be composed on top of Sumo base style.
        no_base_style (:obj:`bool`, optional): Prevent use of sumo base style.
            This can make alternative styles behave more predictably.
        colours (:obj:`dict`, optional): Use custom colours for specific
            element and orbital combinations. Specified as a :obj:`dict` of
            :obj:`dict` of the colours. For example::

                {
                    'Sn': {'s': 'r', 'p': 'b'},
                    'O': {'s': '#000000'}
                }

            The colour can be a hex code, series of rgb value, or any other
            format supported by matplotlib.
        yscale (:obj:`float`, optional): Scaling factor for the y-axis.
        image_format (:obj:`str`, optional): The image file format. Can be any
            format supported by matplotlib, including: png, jpg, pdf, and svg.
            Defaults to pdf.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the image.
        plt (:obj:`matplotlib.pyplot`, optional): A
            :obj:`matplotlib.pyplot` object to use for plotting.
        fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
            a single font, specified as a :obj:`str`, or several fonts,
            specified as a :obj:`list` of :obj:`str`.

    Returns:
        If ``plt`` set then the ``plt`` object will be returned. Otherwise, the
        method will return a :obj:`list` of filenames written to disk.
    """
    if not filenames:
        filenames = find_vasprun_files()
    elif isinstance(filenames, str):
        filenames = [filenames]

    # only load the orbital projects if we definitely need them
    parse_projected = True if projection_selection else False

    # now load all the band structure data and combine using the
    # get_reconstructed_band_structure function from pymatgen
    bandstructures = []
    if code == "vasp":
        for vr_file in filenames:
            vr = BSVasprun(vr_file, parse_projected_eigen=parse_projected)
            bs = vr.get_band_structure(line_mode=True)
            bandstructures.append(bs)
        bs = get_reconstructed_band_structure(bandstructures)
    elif code == "castep":
        for bands_file in filenames:
            cell_file = _replace_ext(bands_file, "cell")
            if os.path.isfile(cell_file):
                logging.info(f"Found cell file {cell_file}...")
            else:
                logging.info(f"Did not find cell file {cell_file}...")
                cell_file = None
            bs = castep_band_structure(bands_file, cell_file=cell_file)
            bandstructures.append(bs)
        bs = get_reconstructed_band_structure(bandstructures)
    elif code == "questaal":
        bnds_file = filenames[0]
        ext = bnds_file.split(".")[-1]
        bnds_folder = os.path.join(bnds_file, os.path.pardir)

        site_file = os.path.abspath(os.path.join(bnds_folder, f"site.{ext}"))

        if os.path.isfile(site_file):
            logging.info("site file found, reading lattice...")
            site_data = QuestaalSite.from_file(site_file)
            bnds_lattice = site_data.structure.lattice
            alat = site_data.alat
        else:
            raise OSError(
                f"Site file {site_file} not found: needed to determine lattice"
            )

        syml_file = os.path.abspath(os.path.join(bnds_folder, f"syml.{ext}"))
        if os.path.isfile(syml_file):
            logging.info("syml file found, reading special-point labels...")
            bnds_labels = labels_from_syml(syml_file)
        else:
            logging.info("syml file not found, band structure lacks labels")
            bnds_labels = {}

        bs = questaal_band_structure(
            bnds_file,
            bnds_lattice,
            alat=alat,
            labels=bnds_labels,
            coords_are_cartesian=cart_coords,
        )

    # currently not supported as it is a pain to make subplots within subplots,
    # although need to check this is still the case
    if "split" in mode and dos_file:
        logging.error(
            "ERROR: Plotting split projected band structure with DOS"
            " not supported.\nPlease use --projected-rgb or "
            "--projected-stacked options."
        )
        sys.exit()

    if projection_selection and mode == "rgb" and len(projection_selection) > 3:
        logging.error(
            "ERROR: RGB projected band structure only "
            "supports up to 3 elements/orbitals."
            "\nUse alternative --mode setting."
        )
        sys.exit()

    # don't save if pyplot object provided
    save_files = False if plt else True
    spin = string_to_spin(spin)  # Convert spin name to pymatgen Spin object

    dos_plotter = None
    dos_opts = None
    if dos_file:
        if code == "vasp":
            dos, pdos = load_dos(
                dos_file,
                elements,
                lm_orbitals,
                atoms,
                gaussian,
                total_only,
                scissor=scissor,
            )
        elif code == "castep":
            if scissor:
                raise ValueError("Scissor not compatabile with CASTEP DOS.")
            pdos_file = None
            if cell_file:
                pdos_file = _replace_ext(cell_file, "pdos_bin")
                if not os.path.isfile(pdos_file):
                    pdos_file = None
                    logging.info(
                        f"PDOS file {pdos_file} does not exist, "
                        "falling back to TDOS."
                    )
                else:
                    logging.info(f"Found PDOS file {pdos_file}")
            else:
                logging.info(f"Cell file {cell_file} does not exist, cannot plot PDOS.")

            dos, pdos = read_castep_dos(
                dos_file,
                pdos_file=pdos_file,
                cell_file=cell_file,
                gaussian=gaussian,
                lm_orbitals=lm_orbitals,
                elements=elements,
                efermi_to_vbm=True,
            )

        dos_plotter = SDOSPlotter(dos, pdos)
        dos_opts = {
            "plot_total": plot_total,
            "legend_cutoff": legend_cutoff,
            "colours": colours,
            "yscale": yscale,
        }

    if scissor:
        bs = bs.apply_scissor(scissor)

    plotter = SBSPlotter(bs)
    if projection_selection:
        plt = plotter.get_projected_plot(
            projection_selection,
            mode=mode,
            normalise=normalise,
            interpolate_factor=interpolate_factor,
            color1=color1,
            color2=color2,
            color3=color3,
            colorspace=colorspace,
            circle_size=circle_size,
            zero_to_efermi=True,
            zero_line=zero_line,
            zero_energy=zero_energy,
            ymin=ymin,
            ymax=ymax,
            height=height,
            width=width,
            vbm_cbm_marker=vbm_cbm_marker,
            ylabel=ylabel,
            plt=plt,
            dos_plotter=dos_plotter,
            dos_options=dos_opts,
            dos_label=dos_label,
            fonts=fonts,
            style=style,
            no_base_style=no_base_style,
            spin=spin,
        )
    else:
        plt = plotter.get_plot(
            zero_to_efermi=True,
            zero_line=zero_line,
            zero_energy=zero_energy,
            ymin=ymin,
            ymax=ymax,
            height=height,
            width=width,
            vbm_cbm_marker=vbm_cbm_marker,
            ylabel=ylabel,
            plt=plt,
            dos_plotter=dos_plotter,
            dos_options=dos_opts,
            dos_label=dos_label,
            fonts=fonts,
            style=style,
            no_base_style=no_base_style,
            spin=spin,
        )

    if save_files:
        basename = f"band.{image_format}"
        filename = f"{prefix}_{basename}" if prefix else basename
        if directory:
            filename = os.path.join(directory, filename)
        plt.savefig(filename, format=image_format, dpi=dpi, bbox_inches="tight")

        written = [filename]
        written += save_data_files(bs, prefix=prefix, directory=directory)
        return written

    else:
        return plt


def find_vasprun_files():
    """Search for vasprun files from the current directory.

    The precedence order for file locations is:

      1. First search for folders named: 'split-0*'
      2. Else, look in the current directory.

    The split folder names should always be zero based, therefore easily
    sortable.
    """
    folders = glob.glob("split-*")
    folders = sorted(folders) if folders else ["."]

    filenames = []
    for fol in folders:
        vr_file = os.path.join(fol, "vasprun.xml")
        vr_file_gz = os.path.join(fol, "vasprun.xml.gz")

        if os.path.exists(vr_file):
            filenames.append(vr_file)
        elif os.path.exists(vr_file_gz):
            filenames.append(vr_file_gz)
        else:
            logging.error(f"ERROR: No vasprun.xml found in {fol}!")
            sys.exit()

    return filenames


def save_data_files(bs, prefix=None, directory=None):
    """Write the band structure data files to disk.

    Args:
        bs (`BandStructureSymmLine`): Calculated band structure.
        prefix (`str`, optional): Prefix for data file.
        directory (`str`, optional): Directory in which to save the data.

    Returns:
        The filename of the written data file.
    """
    filename = f"{prefix}_band.dat" if prefix else "band.dat"
    directory = directory if directory else "."
    filename = os.path.join(directory, filename)

    if bs.is_metal():
        zero = bs.efermi
    else:
        zero = bs.get_vbm()["energy"]

    with open(filename, "w") as f:
        header = "#k-distance eigenvalue[eV]\n"
        f.write(header)

        # write the spin up eigenvalues
        for band in bs.bands[Spin.up]:
            for d, e in zip(bs.distance, band):
                f.write(f"{d:.8f} {e - zero:.8f}\n")
            f.write("\n")

        # calculation is spin polarised, write spin down bands at end of file
        if bs.is_spin_polarized:
            for band in bs.bands[Spin.down]:
                for d, e in zip(bs.distance, band):
                    f.write(f"{d:.8f} {e - zero:.8f}\n")
                f.write("\n")
    return filename


def _el_orb_tuple(string):
    """Parse the element and orbital argument strings.

    The presence of an element without any orbitals means that we want to plot
    all of its orbitals.

    Args:
        string (`str`): The selected elements and orbitals in in the form:
            `"Sn.s.p,O"`.

    Returns:
        A list of tuples specifying which elements/orbitals to plot. The output
        for the above example would be:

            `[('Sn', ('s', 'p')), 'O']`
    """
    el_orbs = []
    for split in string.split(","):
        splits = split.split(".")
        el = splits[0]
        if len(splits) == 1:
            el_orbs.append(el)
        else:
            el_orbs.append((el, tuple(splits[1:])))
    return el_orbs


def _replace_ext(string, new_ext):
    """Replace file extension

    Args:
        string (`str`): The file name with extensions to be replace
        new_ext (`str`): The new extension

    Returns:
        A string with files extension replaced by new_ext
    """
    name, _ = os.path.splitext(string)
    return name + "." + new_ext


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    bandplot is a script to produce publication-ready band
    structure diagrams""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "-f",
        "--filenames",
        default=None,
        nargs="+",
        metavar="F",
        help="one or more vasprun.xml files to plot",
    )
    parser.add_argument(
        "-c",
        "--code",
        default="vasp",
        help="Electronic structure code (default: vasp)." '"questaal" also supported.',
    )
    parser.add_argument(
        "-p", "--prefix", metavar="P", help="prefix for the files generated"
    )
    parser.add_argument(
        "-d", "--directory", metavar="D", help="output directory for files"
    )
    parser.add_argument(
        "-b",
        "--band-edges",
        dest="band_edges",
        action="store_true",
        help="highlight the band edges with markers",
    )
    parser.add_argument(
        "--project",
        default=None,
        metavar="S",
        type=_el_orb_tuple,
        dest="projection_selection",
        help=(
            "select which orbitals to project onto the band "
            'structure (e.g. "Zn.s,Zn.p,O")'
        ),
    )
    parser.add_argument(
        "--mode",
        default="rgb",
        type=str,
        help=("mode for orbital projections (options: rgb, stacked)"),
    )
    parser.add_argument(
        "--normalise",
        default="all",
        type=str,
        help=("how to normalise projections (options: all, select)"),
    )
    parser.add_argument(
        "--interpolate-factor",
        type=int,
        default=4,
        dest="interpolate_factor",
        metavar="N",
        help=("interpolate factor for band structure projections (default: 4)"),
    )
    parser.add_argument(
        "--cartesian",
        action="store_true",
        help="Read cartesian k-point coordinates. This is only"
        " necessary for some Questaal calculations; Vasp "
        "outputs are less ambiguous and this option will "
        "be ignored if --code=vasp.",
    )
    parser.add_argument(
        "--colour1",
        type=str,
        default="#FF0000",
        dest="color1",
        metavar="C",
        help="colour1 for rgb projections (default: red)",
    )
    parser.add_argument(
        "--colour2",
        type=str,
        default="#00FF00",
        dest="color2",
        metavar="C",
        help="colour2 for rgb projections (default: green)",
    )
    parser.add_argument(
        "--colour3",
        type=str,
        default="#0000FF",
        dest="color3",
        metavar="C",
        help="colour3 for rgb projections (default: blue)",
    )
    parser.add_argument(
        "--colourspace",
        type=str,
        default="lab",
        dest="colorspace",
        metavar="C",
        help=(
            "colorspace used for interpolation of for rgb projections (options: "
            "lab[default], rgb, hsv, luvlc, lablch, and xyz)"
        ),
    )
    parser.add_argument(
        "--circle-size",
        type=int,
        default=150,
        dest="circle_size",
        metavar="S",
        help='circle size for "stacked" projections (default: 150)',
    )
    parser.add_argument(
        "--ylabel",
        type=str,
        default="Energy (eV)",
        help="y-axis (i.e. energy) label/units",
    )
    parser.add_argument(
        "--dos-label",
        type=str,
        dest="dos_label",
        default=None,
        help="Axis label for DOS if included",
    )
    parser.add_argument(
        "--dos", default=None, help="path to density of states vasprun.xml"
    )

    parser.add_argument(
        "--zero-line",
        action="store_true",
        dest="zero_line",
        help="Plot horizontal line at energy zero",
    )

    parser.add_argument(
        "--zero-energy",
        type=float,
        dest="zero_energy",
        default=None,
        help=(
            "Reference energy: energy will be shifted to place this energy "
            "at zero. If not specified, zero will be set to the VBM."
        ),
    )
    parser.add_argument(
        "--elements",
        type=_el_orb,
        metavar="E",
        help='elemental orbitals to plot (e.g. "C.s.p,O")',
    )
    parser.add_argument(
        "--orbitals",
        type=_el_orb,
        metavar="O",
        help=("orbitals to split into lm-decomposed " 'contributions (e.g. "Ru.d")'),
    )
    parser.add_argument(
        "--atoms",
        type=_atoms,
        metavar="A",
        help=('atoms to include (e.g. "O.1.2.3,Ru.1.2.3")'),
    )
    parser.add_argument(
        "--spin",
        type=str,
        default=None,
        help=(
            "select only one spin channel for a "
            "spin-polarised calculation "
            "(options: up, 1; down, -1)"
        ),
    )
    parser.add_argument(
        "--scissor",
        type=float,
        default=None,
        dest="scissor",
        help="apply scissor operator",
    )
    parser.add_argument(
        "--total-only",
        action="store_true",
        dest="total_only",
        help="only plot the total density of states",
    )
    parser.add_argument(
        "--no-total",
        action="store_false",
        dest="total",
        help="don't plot the total density of states",
    )
    parser.add_argument(
        "--legend-cutoff",
        type=float,
        default=3,
        dest="legend_cutoff",
        metavar="C",
        help=(
            "cut-off in %% of total DOS that determines if"
            " a line is given a label (default: 3)"
        ),
    )
    parser.add_argument(
        "-g",
        "--gaussian",
        type=float,
        metavar="G",
        help="standard deviation of DOS gaussian broadening",
    )
    parser.add_argument(
        "--scale",
        type=float,
        default=1,
        help="scaling factor for the density of states",
    )
    parser.add_argument(
        "--height", type=float, default=None, help="height of the graph"
    )
    parser.add_argument("--width", type=float, default=None, help="width of the graph")
    parser.add_argument(
        "--ymin", type=float, default=-6.0, help="minimum energy on the y-axis"
    )
    parser.add_argument(
        "--ymax", type=float, default=6.0, help="maximum energy on the y-axis"
    )
    parser.add_argument(
        "--style",
        type=str,
        nargs="+",
        default=None,
        help="matplotlib style specifications",
    )
    parser.add_argument(
        "--no-base-style",
        action="store_true",
        dest="no_base_style",
        help="prevent use of sumo base style",
    )
    parser.add_argument(
        "--config", type=str, default=None, help="colour configuration file"
    )
    parser.add_argument(
        "--format",
        type=str,
        default="pdf",
        dest="image_format",
        metavar="FORMAT",
        help="image file format (options: pdf, svg, jpg, png)",
    )
    parser.add_argument(
        "--dpi", type=int, default=400, help="pixel density for image file"
    )
    parser.add_argument("--font", default=None, help="font to use")
    return parser


def main():
    args = _get_parser().parse_args()
    logging.basicConfig(
        filename="sumo-bandplot.log",
        level=logging.INFO,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    if args.config is None:
        config_path = resource_filename(
            Requirement.parse("sumo"), "sumo/plotting/orbital_colours.conf"
        )
    else:
        config_path = args.config
    colours = configparser.ConfigParser()
    colours.read(os.path.abspath(config_path))

    warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UnicodeWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")

    bandplot(
        filenames=args.filenames,
        code=args.code,
        prefix=args.prefix,
        directory=args.directory,
        vbm_cbm_marker=args.band_edges,
        projection_selection=args.projection_selection,
        mode=args.mode,
        normalise=args.normalise,
        interpolate_factor=args.interpolate_factor,
        cart_coords=args.cartesian,
        scissor=args.scissor,
        color1=args.color1,
        color2=args.color2,
        color3=args.color3,
        colorspace=args.colorspace,
        circle_size=args.circle_size,
        yscale=args.scale,
        ylabel=args.ylabel,
        dos_label=args.dos_label,
        dos_file=args.dos,
        zero_line=args.zero_line,
        zero_energy=args.zero_energy,
        elements=args.elements,
        lm_orbitals=args.orbitals,
        atoms=args.atoms,
        spin=args.spin,
        total_only=args.total_only,
        plot_total=args.total,
        legend_cutoff=args.legend_cutoff,
        gaussian=args.gaussian,
        height=args.height,
        width=args.width,
        ymin=args.ymin,
        style=args.style,
        no_base_style=args.no_base_style,
        ymax=args.ymax,
        colours=colours,
        image_format=args.image_format,
        dpi=args.dpi,
        fonts=args.font,
    )


if __name__ == "__main__":
    main()



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
This module provides a class for plotting electronic band structure diagrams.
"""

import itertools as it
import logging

import numpy as np
from matplotlib import cycler, rcParams
from matplotlib.style import context
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.plotter import BSPlotter
from scipy.interpolate import interp1d

from sumo.electronic_structure.bandstructure import (
    force_branches,
    get_projections_by_branches,
)
from sumo.plotting import (
    colorline,
    draw_themed_line,
    pretty_plot,
    pretty_subplot,
    styled_plot,
    sumo_base_style,
    sumo_bs_style,
)

label_size = 22


class SBSPlotter(BSPlotter):
    """Class for plotting electronic band structures.

    This class is similar to the pymatgen
    :obj:`pymatgen.electronic_structure.plotter.BSPlotter` class but overrides
    some methods to generate prettier plots.

    Additional functionality, such as projected band structure plots are
    available.

    Args:
        bs (:obj:`~pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`):
            The band structure.
    """

    def __init__(self, bs):
        if isinstance(bs, list):
            bs = [force_branches(b) for b in bs]
        else:
            bs = force_branches(bs)

        BSPlotter.__init__(self, bs)

        # old versions of pymatgen only support a single band structure
        if isinstance(self._bs, list):
            self.bs = self._bs[0]
            self.nbands = self._nb_bands[0]
        else:
            self.bs = self._bs
            self.nbands = self._nb_bands

    @staticmethod
    def _reset_zero_energy(bs_plot_data, zero_energy=0.0):
        """Modify reference energy of data from bs_plot_data

        This method is defined in Pymatgen and obtains plotting data with zero
        set to 0. (if zero_to_efermi=False), self.efermi (if
        zero_to_efermi=True and metallic) or the absolute VBM energy
        (zero_to_efermi=False and insulating).

        Sumo allows other reference energies to be specified (e.g. from
        sc-fermi); conveniently the data dictionary from Pymatgen contains a
        "zero_energy" field so the data can always be related to absolute (DFT)
        values.

        Args:
            bs_plot_data (:obj:`dict`): data dictionary from
                pymatgen.electronicstructure.plotter.BSPlotter.bs_plot_data()
            zero_energy (:obj:`float`, Optional): New reference energy

        Returns:
            dict in format of bs_plot_data.
        """

        shifted_data = {}
        energy_shift = bs_plot_data["zero_energy"] - zero_energy

        for key, value in bs_plot_data.items():
            if key in ("vbm", "cbm"):
                shifted_data[key] = [(pt[0], pt[1] + energy_shift) for pt in value]
            elif key == "energy":
                shifted_data["energy"] = {}
                for spin, energies in value.items():
                    shifted_data["energy"][spin] = [
                        array + energy_shift for array in energies
                    ]
            elif key == "zero_energy":
                shifted_data[key] = zero_energy
            else:
                shifted_data[key] = value
        return shifted_data

    @styled_plot(sumo_base_style, sumo_bs_style)
    def get_plot(
        self,
        zero_to_efermi=True,
        zero_line=False,
        zero_energy=None,
        ymin=-6.0,
        ymax=6.0,
        width=None,
        height=None,
        vbm_cbm_marker=False,
        ylabel="Energy (eV)",
        dpi=None,
        plt=None,
        plot_dos_legend=True,
        dos_plotter=None,
        dos_options=None,
        dos_label=None,
        dos_aspect=3,
        aspect=None,
        spin=None,
        fonts=None,
        style=None,
        no_base_style=False,
    ):
        """Get a :obj:`matplotlib.pyplot` object of the band structure.

        If the system is spin polarised, and no spin has been specified, orange
        lines are spin up, dashed blue lines are spin down. For metals, all
        bands are coloured blue. For semiconductors, blue lines indicate
        valence bands and orange lines indicates conduction bands.

        Args:
            zero_to_efermi (:obj:`bool`): Shift the plot such that the
                Fermi energy of the band structure data is plotted at 0 eV.
                Note that this "efermi" is for the benefit of Pymatgen and may
                not reflect the actual Fermi level; in Sumo it has usually
                already been shifted to the VBM.
            zero_line (:obj:`bool`, optional): Draw a horizontal line at zero
            zero_energy (:obj:`float`, optional): Zero energy reference. (If
                unset, defaults to VBM or Fermi Energy as appropriate.)
            ymin (:obj:`float`, optional): The minimum energy on the y-axis.
            ymax (:obj:`float`, optional): The maximum energy on the y-axis.
            width (:obj:`float`, optional): The width of the plot.
            height (:obj:`float`, optional): The height of the plot.
            vbm_cbm_marker (:obj:`bool`, optional): Plot markers to indicate
                the VBM and CBM locations.
            ylabel (:obj:`str`, optional): y-axis (i.e. energy) label/units
            dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
                the image.
            plt (:obj:`matplotlib.pyplot`, optional): A
                :obj:`matplotlib.pyplot` object to use for plotting.
            dos_plotter (:obj:`~sumo.plotting.dos_plotter.SDOSPlotter`, \
                optional): Plot the density of states alongside the band
                structure. This should be a
                :obj:`~sumo.plotting.dos_plotter.SDOSPlotter` object
                initialised with the data to plot.
            dos_options (:obj:`dict`, optional): The options for density of
                states plotting. This should be formatted as a :obj:`dict`
                containing any of the following keys:

                    "yscale" (:obj:`float`)
                        Scaling factor for the y-axis.
                    "xmin" (:obj:`float`)
                        The minimum energy to mask the energy and density of
                        states data (reduces plotting load).
                    "xmax" (:obj:`float`)
                        The maximum energy to mask the energy and density of
                        states data (reduces plotting load).
                    "colours" (:obj:`dict`)
                        Use custom colours for specific element and orbital
                        combinations. Specified as a :obj:`dict` of
                        :obj:`dict` of the colours. For example::

                            {
                                'Sn': {'s': 'r', 'p': 'b'},
                                'O': {'s': '#000000'}
                            }

                        The colour can be a hex code, series of rgb value, or
                        any other format supported by matplotlib.
                    "plot_total" (:obj:`bool`)
                        Plot the total density of states. Defaults to ``True``.
                    "legend_cutoff" (:obj:`float`)
                        The cut-off (in % of the maximum density of states
                        within the plotting range) for an elemental orbital to
                        be labelled in the legend. This prevents the legend
                        from containing labels for orbitals that have very
                        little contribution in the plotting range.
                    "subplot" (:obj:`bool`)
                        Plot the density of states for each element on separate
                        subplots. Defaults to ``False``.

            dos_label (:obj:`str`, optional): DOS axis label/units
            dos_aspect (:obj:`float`, optional): Aspect ratio for the band
                structure and density of states subplot. For example,
                ``dos_aspect = 3``, results in a ratio of 3:1, for the band
                structure:dos plots.
            plot_dos_legend (:obj:`bool`): Whether to plot the dos legend.
            aspect (:obj:`float`, optional): The aspect ratio of the band
                structure plot. By default the dimensions of the figure size
                are used to determine the aspect ratio. Set to ``1`` to force
                the plot to be square.
            spin (:obj:`Spin`, optional): Plot a spin-polarised band structure,
                "up" or "1" for spin up only, "down" or "-1" for spin down only.
                Defaults to ``None``.
            fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
                a single font, specified as a :obj:`str`, or several fonts,
                specified as a :obj:`list` of :obj:`str`.
            style (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
                style specifications, to be composed on top of Sumo base
                style.
                no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.
            no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.

        Returns:
            :obj:`matplotlib.pyplot`: The electronic band structure plot.
        """
        if dos_plotter:
            plt = pretty_subplot(
                1,
                2,
                width=width,
                height=height,
                sharex=False,
                dpi=dpi,
                plt=plt,
                gridspec_kw={"width_ratios": [dos_aspect, 1], "wspace": 0},
            )
            ax = plt.gcf().axes[0]
        else:
            plt = pretty_plot(width=width, height=height, dpi=dpi, plt=plt)
            ax = plt.gca()

        data = self.bs_plot_data(zero_to_efermi=True)
        if zero_energy is not None:
            data = self._reset_zero_energy(data, zero_energy=zero_energy)

        dists = data["distances"]
        eners = data["energy"]

        if spin is not None and not self.bs.is_spin_polarized:
            raise ValueError(
                "Spin-selection only possible with spin-polarised "
                "calculation results"
            )
        elif self.bs.is_metal() or (self.bs.is_spin_polarized and not spin):
            # if metal or spin polarized and spin not specified
            is_vb = [True]
        elif spin:
            # not metal, spin-polarized and spin is set
            is_vb = self.bs.bands[spin] <= self.bs.get_vbm()["energy"]
        else:
            # not metal, not spin polarized and therefore spin not set
            is_vb = self.bs.bands[Spin.up] <= self.bs.get_vbm()["energy"]

        # nd is branch index, nb is band index, nk is kpoint index
        for nd, nb in it.product(range(len(data["distances"])), range(self.nbands)):
            e = (
                eners[str(spin)][nd][nb]
                if spin is not None
                else eners[str(Spin.up)][nd][nb]
            )

            # For closed-shell calculations with a bandgap, colour valence
            # bands blue (C0) and conduction bands orange (C1)
            #
            # For closed-shell calculations with no bandgap, colour with C0
            #
            # For spin-polarized calculations, colour spin up channel with C1
            # and overlay with C0 (dashed) spin down channel

            if self.bs.is_spin_polarized and spin is None:
                c = "C1"
            elif self.bs.is_metal() or np.all(is_vb[nb]):
                c = "C0"
            else:
                c = "C1"

            ax.plot(dists[nd], e, ls="-", c=c, zorder=1)

        # Plot second spin channel if it exists and no spin selected
        if self.bs.is_spin_polarized and spin is None:
            for nd, nb in it.product(range(len(data["distances"])), range(self.nbands)):
                e = eners[str(Spin.down)][nd][nb]
                ax.plot(dists[nd], e, c="C0", linestyle="--", zorder=2)

        self._maketicks(ax, ylabel=ylabel)

        self._makeplot(
            ax,
            plt.gcf(),
            data,
            zero_to_efermi=zero_to_efermi,
            zero_line=zero_line,
            vbm_cbm_marker=vbm_cbm_marker,
            width=width,
            height=height,
            ymin=ymin,
            ymax=ymax,
            dos_plotter=dos_plotter,
            dos_options=dos_options,
            plot_dos_legend=plot_dos_legend,
            dos_label=dos_label,
            aspect=aspect,
            spin=spin,
        )
        return plt

    @styled_plot(sumo_base_style, sumo_bs_style)
    def get_projected_plot(
        self,
        selection,
        mode="rgb",
        normalise="all",
        interpolate_factor=4,
        color1="#FF0000",
        color2="#00FF00",
        color3="#0000FF",
        colorspace="lab",
        circle_size=150,
        projection_cutoff=0.001,
        zero_energy=None,
        zero_to_efermi=True,
        zero_line=False,
        ymin=-6.0,
        ymax=6.0,
        width=None,
        height=None,
        vbm_cbm_marker=False,
        ylabel="Energy (eV)",
        dpi=400,
        plt=None,
        dos_plotter=None,
        dos_options=None,
        dos_label=None,
        plot_dos_legend=True,
        dos_aspect=3,
        aspect=None,
        fonts=None,
        style=None,
        no_base_style=False,
        spin=None,
    ):
        """Get a :obj:`matplotlib.pyplot` of the projected band structure.

        If the system is spin polarised, no spin has been specified and
        ``mode = 'rgb'`` spin up and spin down bands are differentiated by
        solid and dashed lines, respectively.
        For the other modes, spin up and spin down are plotted separately.

        Args:
            selection (list): A list of :obj:`tuple` or :obj:`string`
                identifying which elements and orbitals to project on to the
                band structure. These can be specified by both element and
                orbital, for example, the following will project the Bi s, p
                and S p orbitals::

                    [('Bi', 's'), ('Bi', 'p'), ('S', 'p')]

                If just the element is specified then all the orbitals of
                that element are combined. For example, to sum all the S
                orbitals::

                    [('Bi', 's'), ('Bi', 'p'), 'S']

                You can also choose to sum particular orbitals by supplying a
                :obj:`tuple` of orbitals. For example, to sum the S s, p, and
                d orbitals into a single projection::

                  [('Bi', 's'), ('Bi', 'p'), ('S', ('s', 'p', 'd'))]

                If ``mode = 'rgb'``, a maximum of 3 orbital/element
                combinations can be plotted simultaneously (one for red, green
                and blue), otherwise an unlimited number of elements/orbitals
                can be selected.
            mode (:obj:`str`, optional): Type of projected band structure to
                plot. Options are:

                    "rgb"
                        The band structure line color depends on the character
                        of the band. Each element/orbital contributes either
                        red, green or blue with the corresponding line colour a
                        mixture of all three colours. This mode only supports
                        up to 3 elements/orbitals combinations. The order of
                        the ``selection`` :obj:`tuple` determines which colour
                        is used for each selection.
                    "stacked"
                        The element/orbital contributions are drawn as a
                        series of stacked circles, with the colour depending on
                        the composition of the band. The size of the circles
                        can be scaled using the ``circle_size`` option.

            normalise (:obj:`str`, optional): Normalisation the projections.
                Options are:

                  * ``'all'``: Projections normalised against the sum of all
                       other projections.
                  * ``'select'``: Projections normalised against the sum of the
                       selected projections.
                  * ``None``: No normalisation performed.

            interpolate_factor (:obj:`int`, optional): The factor by which to
                interpolate the band structure (necessary to make smooth
                lines). A larger number indicates greater interpolation.
            circle_size (:obj:`float`, optional): The area of the circles used
                when ``mode = 'stacked'``.
            color1 (str): A color specified in any way supported by matplotlib. Used
                when ``mode = 'rgb'``.
            color2 (str): A color specified in any way supported by matplotlib. Used
                when ``mode = 'rgb'``.
            color3 (str): A color specified in any way supported by matplotlib. Used
                when ``mode = 'rgb'``.
            colorspace (str): The colorspace in which to perform the interpolation. The
                allowed values are rgb, hsv, lab, luvlc, lablch, and xyz. Used
                when ``mode = 'rgb'``.
            projection_cutoff (:obj:`float`): Don't plot projections with
                intensities below this number. This option is useful for
                stacked plots, where small projections clutter the plot.
            zero_to_efermi (:obj:`bool`): Normalise the plot such that the
                valence band maximum is set as 0 eV.
            zero_line (:obj:`bool`, optional): Draw a horizontal line at zero
            zero_energy (:obj:`float`, optional): Zero energy reference. (If
                unset, defaults to VBM or Fermi Energy as appropriate.)
            ymin (:obj:`float`, optional): The minimum energy on the y-axis.
            ymax (:obj:`float`, optional): The maximum energy on the y-axis.
            width (:obj:`float`, optional): The width of the plot.
            height (:obj:`float`, optional): The height of the plot.
            vbm_cbm_marker (:obj:`bool`, optional): Plot markers to indicate
                the VBM and CBM locations.
            ylabel (:obj:`str`, optional): y-axis (i.e. energy) label/units
            dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
                the image.
            plt (:obj:`matplotlib.pyplot`, optional): A
                :obj:`matplotlib.pyplot` object to use for plotting.
            dos_plotter (:obj:`~sumo.plotting.dos_plotter.SDOSPlotter`, \
                optional): Plot the density of states alongside the band
                structure. This should be a
                :obj:`~sumo.plotting.dos_plotter.SDOSPlotter` object
                initialised with the data to plot.
            dos_options (:obj:`dict`, optional): The options for density of
                states plotting. This should be formatted as a :obj:`dict`
                containing any of the following keys:

                    "yscale" (:obj:`float`)
                        Scaling factor for the y-axis.
                    "xmin" (:obj:`float`)
                        The minimum energy to mask the energy and density of
                        states data (reduces plotting load).
                    "xmax" (:obj:`float`)
                        The maximum energy to mask the energy and density of
                        states data (reduces plotting load).
                    "colours" (:obj:`dict`)
                        Use custom colours for specific element and orbital
                        combinations. Specified as a :obj:`dict` of
                        :obj:`dict` of the colours. For example::

                           {
                                'Sn': {'s': 'r', 'p': 'b'},
                                'O': {'s': '#000000'}
                            }

                        The colour can be a hex code, series of rgb value, or
                        any other format supported by matplotlib.
                    "plot_total" (:obj:`bool`)
                        Plot the total density of states. Defaults to ``True``.
                    "legend_cutoff" (:obj:`float`)
                        The cut-off (in % of the maximum density of states
                        within the plotting range) for an elemental orbital to
                        be labelled in the legend. This prevents the legend
                        from containing labels for orbitals that have very
                        little contribution in the plotting range.
                    "subplot" (:obj:`bool`)
                        Plot the density of states for each element on separate
                        subplots. Defaults to ``False``.

            dos_label (:obj:`str`, optional): DOS axis label/units
            plot_dos_legend (:obj:`bool`): Whether to plot the dos legend.
            dos_aspect (:obj:`float`, optional): Aspect ratio for the band
                structure and density of states subplot. For example,
                ``dos_aspect = 3``, results in a ratio of 3:1, for the band
                structure:dos plots.
            aspect (:obj:`float`, optional): The aspect ratio of the band
                structure plot. By default the dimensions of the figure size
                are used to determine the aspect ratio. Set to ``1`` to force
                the plot to be square.
            fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
                a single font, specified as a :obj:`str`, or several fonts,
                specified as a :obj:`list` of :obj:`str`.
            style (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
                style specifications, to be composed on top of Sumo base
                style.
            no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.
            spin (:obj:`Spin`, optional): Plot a spin-polarised band structure,
                "up" or "1" for spin up only, "down" or "-1" for spin down
                only. Defaults to ``None``.

        Returns:
            :obj:`matplotlib.pyplot`: The projected electronic band structure
            plot.
        """
        if mode == "rgb" and len(selection) > 3:
            raise ValueError("Too many elements/orbitals specified (max 3)")
        elif mode == "solo" and dos_plotter:
            raise ValueError("Solo mode plotting with DOS not supported")

        if dos_plotter:
            plt = pretty_subplot(
                1,
                2,
                width,
                height,
                sharex=False,
                dpi=dpi,
                plt=plt,
                gridspec_kw={"width_ratios": [dos_aspect, 1], "wspace": 0},
            )
            ax = plt.gcf().axes[0]
        else:
            plt = pretty_plot(width, height, dpi=dpi, plt=plt)
            ax = plt.gca()

        data = self.bs_plot_data(zero_to_efermi=zero_to_efermi)
        if zero_energy is not None:
            data = self._reset_zero_energy(data, zero_energy=zero_energy)

        nbranches = len(data["distances"])

        # Ensure we do spin up first, then spin down
        spins = sorted(self.bs.bands.keys(), key=lambda s: -s.value)
        if spin is not None and len(spins) == 1:
            raise ValueError(
                "Spin-selection only possible with spin-polarised "
                "calculation results"
            )

        if spin is Spin.up:
            spins = [spins[0]]
        elif spin is Spin.down:
            spins = [spins[1]]

        proj = get_projections_by_branches(self.bs, selection, normalise=normalise)

        # nd is branch index
        for spin, nd in it.product(spins, range(nbranches)):

            # mask data to reduce plotting load
            bands = np.array(data["energy"][str(spin)][nd])
            mask = np.where(
                np.any(bands > ymin - 0.05, axis=1)
                & np.any(bands < ymax + 0.05, axis=1)
            )
            distances = data["distances"][nd]
            bands = bands[mask]
            weights = [proj[nd][i][spin][mask] for i in range(len(selection))]

            if len(distances) > 2:  # Only interpolate if it makes sense to do so
                # interpolate band structure to improve smoothness
                temp_dists = np.linspace(
                    distances[0], distances[-1], len(distances) * interpolate_factor
                )
                bands = interp1d(
                    distances,
                    bands,
                    axis=1,
                    bounds_error=False,
                    fill_value="extrapolate",
                )(temp_dists)
                weights = interp1d(
                    distances,
                    weights,
                    axis=2,
                    bounds_error=False,
                    fill_value="extrapolate",
                )(temp_dists)
                distances = temp_dists

            else:  # change from list to array if we skipped the scipy interpolation
                weights = np.array(weights)
                bands = np.array(bands)
                distances = np.array(distances)

            # sometimes VASP produces very small negative weights
            weights[weights < 0] = 0

            if mode == "rgb":

                # colours aren't used now but needed later for legend
                colours = [color1, color2, color3]

                # if only two orbitals then just use red and blue
                if len(weights) == 2:
                    weights = np.insert(weights, 1, np.zeros(weights[0].shape), axis=0)
                    colours = ["#ff0000", "#0000ff"]

                ls = "-" if spin == Spin.up else "--"
                lc = colorline(
                    distances,
                    bands,
                    weights.transpose((1, 2, 0)),
                    color1=color1,
                    color2=color2,
                    color3=color3,
                    colorspace=colorspace,
                    linestyles=ls,
                    linewidth=(rcParams["lines.linewidth"] * 1.25),
                )
                ax.add_collection(lc)

            elif mode == "stacked":
                # TODO: Handle spin

                # use some nice custom colours first, then default colours
                colours = ["#3952A3", "#FAA41A", "#67BC47", "#6ECCDD", "#ED2025"]
                colour_series = rcParams["axes.prop_cycle"].by_key()["color"]
                colours.extend(colour_series)

                # very small circles look crap
                weights[weights < projection_cutoff] = 0

                distances = list(distances) * len(bands)
                bands = bands.flatten()
                zorders = range(-len(weights), 0)
                for w, c, z in zip(weights, colours, zorders):
                    ax.scatter(
                        distances,
                        bands,
                        c=c,
                        s=circle_size * w**2,
                        zorder=z,
                        rasterized=True,
                    )

        # plot the legend
        for c, spec in zip(colours, selection):
            if isinstance(spec, str):
                label = spec
            else:
                label = f"{spec[0]} ({' + '.join(spec[1])})"
            ax.scatter([-10000], [-10000], c=c, s=50, label=label, edgecolors="none")

        if dos_plotter:
            loc = 1
            anchor_point = (-0.2, 1)
        else:
            loc = 2
            anchor_point = (0.95, 1)

        ax.legend(
            bbox_to_anchor=anchor_point,
            loc=loc,
            frameon=False,
            handletextpad=0.1,
            scatterpoints=1,
        )

        # finish and tidy plot
        self._maketicks(ax, ylabel=ylabel)
        self._makeplot(
            ax,
            plt.gcf(),
            data,
            zero_to_efermi=zero_to_efermi,
            zero_line=zero_line,
            vbm_cbm_marker=vbm_cbm_marker,
            width=width,
            height=height,
            ymin=ymin,
            ymax=ymax,
            dos_plotter=dos_plotter,
            dos_options=dos_options,
            dos_label=dos_label,
            plot_dos_legend=plot_dos_legend,
            aspect=aspect,
            spin=spin,
        )
        return plt

    def _makeplot(
        self,
        ax,
        fig,
        data,
        zero_to_efermi=True,
        zero_line=False,
        vbm_cbm_marker=False,
        ymin=-6.0,
        ymax=6.0,
        height=None,
        width=None,
        dos_plotter=None,
        dos_options=None,
        dos_label=None,
        plot_dos_legend=True,
        aspect=None,
        spin=None,
    ):
        """Tidy the band structure & add the density of states if required."""
        if zero_line:
            draw_themed_line(0, ax)

        # Draw line at Fermi level if not zeroing to e-Fermi
        elif not zero_to_efermi:
            ef = self.bs.efermi
            draw_themed_line(ef, ax)

        # set x and y limits
        ax.set_xlim(0, data["distances"][-1][-1])
        if self.bs.is_metal() and not zero_to_efermi:
            ax.set_ylim(self.bs.efermi + ymin, self.bs.efermi + ymax)
        else:
            ax.set_ylim(ymin, ymax)

        if vbm_cbm_marker:
            for cbm in data["cbm"]:
                ax.scatter(cbm[0], cbm[1], color="C2", marker="o", s=100)
            for vbm in data["vbm"]:
                ax.scatter(vbm[0], vbm[1], color="C3", marker="o", s=100)

        if dos_plotter:
            ax = fig.axes[1]

            if not dos_options:
                dos_options = {}

            dos_options.update(
                {
                    "xmin": ymin,
                    "xmax": ymax,
                    "zero_energy": data["zero_energy"],
                    "zero_to_efermi": False,
                }
            )
            self._makedos(
                ax,
                dos_plotter,
                dos_options,
                dos_label=dos_label,
                plot_legend=plot_dos_legend,
                zero_line=zero_line,
                spin=spin,
            )
        else:
            # keep correct aspect ratio for axes based on canvas size
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()
            if width is None:
                width = rcParams["figure.figsize"][0]
            if height is None:
                height = rcParams["figure.figsize"][1]

            if aspect is not False:
                if not aspect:
                    aspect = height / width

                ax.set_aspect(aspect * ((x1 - x0) / (y1 - y0)))

    @staticmethod
    def _makedos(
        ax,
        dos_plotter,
        dos_options,
        dos_label=None,
        plot_legend=True,
        zero_line=False,
        spin=None,
    ):
        """This is basically the same as the SDOSPlotter get_plot function."""

        # don't use first 4 colours; these are the band structure line colours
        cycle = cycler("color", rcParams["axes.prop_cycle"].by_key()["color"][4:])
        with context({"axes.prop_cycle": cycle}):
            plot_data = dos_plotter.dos_plot_data(**dos_options)

        mask = plot_data["mask"]
        energies = plot_data["energies"][mask]
        lines = plot_data["lines"]
        if spin is None:
            spins = [Spin.up] if len(lines[0][0]["dens"]) == 1 else [Spin.up, Spin.down]
        elif isinstance(spin, Spin):
            spins = [spin]
        else:
            spins = spin

        # disable y ticks for DOS panel
        ax.tick_params(axis="y", which="both", right=False)

        for line_set in plot_data["lines"]:
            for line, spin in it.product(line_set, spins):
                if spin == Spin.up or len(spins) == 1:
                    label = line["label"]
                    densities = line["dens"][spin][mask]
                else:
                    label = ""
                    densities = -line["dens"][spin][mask]
                ax.fill_betweenx(
                    energies,
                    densities,
                    0,
                    lw=0,
                    facecolor=line["colour"],
                    alpha=line["alpha"],
                )
                ax.plot(densities, energies, label=label, color=line["colour"])

        # x and y axis reversed versus normal dos plotting
        ax.set_ylim(dos_options["xmin"], dos_options["xmax"])
        if len(spins) == 1:
            ax.set_xlim(0, plot_data["ymax"])
        else:
            ax.set_xlim(plot_data["ymin"], plot_data["ymax"])

        if zero_line:
            draw_themed_line(0, ax)

        if dos_label is not None:
            ax.set_xlabel(dos_label)

        ax.set_xticklabels([])
        if plot_legend:
            ax.legend(loc=2, frameon=False, ncol=1, bbox_to_anchor=(1.0, 1.0))

    @staticmethod
    def _sanitise_label(label):
        """Implement label hacks: Hide trailing @, remove label with leading @"""

        import re

        if re.match("^@.*$", label):
            return None
        else:
            return re.sub("@+$", "", label)

    @classmethod
    def _sanitise_label_group(cls, labelgroup):
        r"""Implement label hacks: Hide trailing @, remove label with leading @

        Labels split with $\mid$ symbol will be treated for each part.
        """

        if r"$\mid$" in labelgroup:
            label_components = labelgroup.split(r"$\mid$")
            good_labels = [
                i for i in map(cls._sanitise_label, label_components) if i is not None
            ]
            if len(good_labels) == 0:
                return None
            else:
                return r"$\mid$".join(good_labels)
        else:
            return cls._sanitise_label(labelgroup)

    def _maketicks(self, ax, ylabel="Energy (eV)"):
        """Utility method to add tick marks to a band structure."""
        # set y-ticks
        ax.yaxis.set_major_locator(MaxNLocator(6))
        ax.yaxis.set_minor_locator(AutoMinorLocator(2))

        # set x-ticks; only plot the unique tick labels
        ticks = self.get_ticks()
        unique_d = []
        unique_l = []
        if ticks["distance"]:
            temp_ticks = list(zip(ticks["distance"], ticks["label"]))
            unique_d.append(temp_ticks[0][0])
            unique_l.append(temp_ticks[0][1])
            for i in range(1, len(temp_ticks)):
                # Hide labels marked with @
                if "@" in temp_ticks[i][1]:
                    # If a branch connection, check all parts of label
                    if r"$\mid$" in temp_ticks[i][1]:
                        label_components = temp_ticks[i][1].split(r"$\mid$")
                        good_labels = [i for i in label_components if i[0] != "@"]
                        if len(good_labels) == 0:
                            continue
                        else:
                            temp_ticks[i] = (
                                temp_ticks[i][0],
                                r"$\mid$".join(good_labels),
                            )
                    # If a single label, check first character
                    elif temp_ticks[i][1][0] == "@":
                        continue

                # Append label to sequence if it is not same as predecessor
                if unique_l[-1] != temp_ticks[i][1]:
                    unique_d.append(temp_ticks[i][0])
                    unique_l.append(temp_ticks[i][1])

        logging.info("Label positions:")
        for dist, label in list(zip(unique_d, unique_l)):
            logging.info(f"\t{dist:.4f}: {label}")

        ax.set_xticks(unique_d)
        ax.set_xticklabels(unique_l)
        ax.xaxis.grid(True)
        ax.set_ylabel(ylabel)
